import pytest
from testbook import testbook

from tests.conftest import REPO_ROOT

optuna = pytest.importorskip("optuna")


@testbook(
    REPO_ROOT / "examples/usecases/retrieval-with-hyperparameter-optimization.ipynb", execute=False
)
def test_usecase_pretrained_embeddings(tb):
    tb.inject(
        """
        from merlin.datasets.synthetic import generate_data
        ds = generate_data('transactions', 10000)
        df = ds.compute()
        from datetime import datetime, timedelta
        import random
        def generate_date():
            date = datetime.today()
            if random.randint(0, 1):
                date -= timedelta(days=7)
            return date
        t_dat = [generate_date() for _ in range(df.shape[0])]
        df['t_dat'] = t_dat
        df.to_csv('transactions_train.csv')
        """
    )
    tb.cells[
        27
    ].source = (
        "search_space = {'learning_rate': [1e-3], 'num_epochs': [1, 2], 'embedding_dim': [16]}"
    )
    tb.execute()
    study = tb.ref("study")
    assert set(study.best_params.keys()) == set(["learning_rate", "num_epochs", "embedding_dim"])



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import pathlib
from typing import Optional, Union

from merlin.models.api import MerlinModel
from merlin.models.utils.schema_utils import schema_to_tensorflow_metadata_json
from merlin.schema import Schema

_MERLIN_METADATA_DIR_NAME = ".merlin"


def save_merlin_metadata(
    export_path: Union[str, os.PathLike],
    model: MerlinModel,
    input_schema: Optional[Schema],
    output_schema: Optional[Schema],
) -> None:
    """Saves data to Merlin Metadata Directory."""
    export_path = pathlib.Path(export_path)
    merlin_metadata_dir = export_path / _MERLIN_METADATA_DIR_NAME
    merlin_metadata_dir.mkdir(exist_ok=True)

    if input_schema is not None:
        schema_to_tensorflow_metadata_json(
            input_schema,
            merlin_metadata_dir / "input_schema.json",
        )
    if output_schema is not None:
        schema_to_tensorflow_metadata_json(
            output_schema,
            merlin_metadata_dir / "output_schema.json",
        )



import pytest
from testbook import testbook

from tests.conftest import REPO_ROOT

pytest.importorskip("transformers")


@testbook(
    REPO_ROOT / "examples/usecases/transformers-next-item-prediction.ipynb",
    timeout=180,
    execute=False,
)
def test_next_item_prediction(tb):
    tb.inject(
        """
        import os, random
        from datetime import datetime, timedelta
        from merlin.datasets.synthetic import generate_data
        ds = generate_data('booking.com-raw', 10000)
        df = ds.compute()
        def generate_date():
            date = datetime.today()
            if random.randint(0, 1):
                date -= timedelta(days=7)
            return date
        df['checkin'] = [generate_date() for _ in range(df.shape[0])]
        df['checkout'] = [generate_date() for _ in range(df.shape[0])]
        df.to_csv('/tmp/train_set.csv')
        """
    )
    tb.cells[4].source = tb.cells[4].source.replace("get_booking('/workspace/data')", "")
    tb.cells[4].source = tb.cells[4].source.replace(
        "read_csv('/workspace/data/train_set.csv'", "read_csv('/tmp/train_set.csv'"
    )
    tb.cells[28].source = tb.cells[28].source.replace("d_model=64", "d_model=40")
    tb.cells[30].source = tb.cells[30].source.replace("epochs=5", "epochs=1")
    tb.execute()



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import List, Optional, Union

import tensorflow as tf

from merlin.models.tf.core.base import Block
from merlin.models.tf.core.combinators import ResidualBlock, SequentialBlock, TabularAggregationType
from merlin.models.tf.core.tabular import Filter, tabular_aggregation_registry
from merlin.models.tf.utils.tf_utils import (
    maybe_deserialize_keras_objects,
    maybe_serialize_keras_objects,
)
from merlin.models.utils.misc_utils import filter_kwargs
from merlin.schema import Schema, Tags

InitializerType = Union[str, tf.keras.initializers.Initializer]
RegularizerType = Union[str, tf.keras.regularizers.Regularizer]


def MLPBlock(
    dimensions: List[int],
    activation: Union[str, List[str]] = "relu",
    use_bias: bool = True,
    kernel_initializer: InitializerType = "glorot_uniform",
    bias_initializer: InitializerType = "zeros",
    kernel_regularizer: Optional[RegularizerType] = None,
    bias_regularizer: Optional[RegularizerType] = None,
    activity_regularizer: Optional[RegularizerType] = None,
    dropout: Optional[float] = None,
    normalization: Optional[Union[str, tf.keras.layers.Layer]] = None,
    filter: Optional[Union[Schema, Tags, List[str], "Filter"]] = None,
    no_activation_last_layer: bool = False,
    block_name: str = "MLPBlock",
    **kwargs,
) -> SequentialBlock:
    """
    A block that applies a multi-layer perceptron to the input.

    Example usage::
        mlp = ml.InputBlock(schema).connect(ml.MLPBlock([64, 32]))

    Parameters
    ----------
    dimensions: List[int]
        The number of units in each layer of the MLP.
    activation: str
        The activation function to use.
    use_bias: bool
        Whether to use a bias in the MLP.
    kernel_initializer: InitializerType
        Initializer for the kernel weights matrix. Defaults to "glorot_uniform".
    bias_initializer: InitializerType
        Initializer for the bias vector. Default to "zeros".
    kernel_regularizer: Optional[RegularizerType]
        Regularizer function applied to the kernel weights matrix. Default to None.
    bias_regularizer: Optional[RegularizerType]
        Regularizer function applied to the bias vector.  Default to None.
    activity_regularizer: Optional[RegularizerType]
        Regularizer function applied to the output of the layer (its "activation").
        Default to None.
    dropout: float
        The dropout rate to use.
    normalization: str or Layer
        The normalization layer to use.
    filter: Schema, Tag, List[str], or Filter
        The filter to apply to the inputs of the MLP.
    no_activation_last_layer: bool
        Ensures that no activation function (i.e. 'linear') or droptout is used in the
        output of the last MLP layer
    block_name: str
        The name of the block.
    """

    if isinstance(activation, list) and len(activation) != len(dimensions):
        raise ValueError(
            f"Activation and Dimensions length mismatch. \
        Activation length: {len(activation)}, Dimensions length: {len(dimensions)}"
        )

    block_layers = []

    for idx, dim in enumerate(dimensions):
        dropout_layer = None
        activation_idx = activation if isinstance(activation, str) else activation[idx]
        if no_activation_last_layer and idx == len(dimensions) - 1:
            activation_idx = "linear"
        else:
            if dropout:
                if activation_idx in ["selu", tf.keras.activations.selu]:
                    # Best practice for SeLU. It is also recommended
                    # kernel_initializer="lecun_normal"
                    dropout_layer = tf.keras.layers.AlphaDropout(dropout)
                else:
                    dropout_layer = tf.keras.layers.Dropout(dropout)

        block_layers.append(
            _Dense(
                dim,
                activation=activation_idx,
                use_bias=use_bias,
                kernel_initializer=kernel_initializer,
                bias_initializer=bias_initializer,
                kernel_regularizer=kernel_regularizer,
                bias_regularizer=bias_regularizer,
                activity_regularizer=activity_regularizer,
            )
        )
        if dropout_layer:
            block_layers.append(dropout_layer)

        if normalization:
            if normalization == "batch_norm":
                block_layers.append(tf.keras.layers.BatchNormalization())
            elif isinstance(normalization, tf.keras.layers.Layer):
                block_layers.append(normalization)
            else:
                raise ValueError("Normalization needs to be an instance `Layer` or " "`batch_norm`")

    return SequentialBlock(block_layers, filter=filter, block_name=block_name, **kwargs)


def DenseResidualBlock(
    low_rank_dim: Optional[int] = None,
    activation: Optional[Union[str, tf.keras.layers.Layer]] = "relu",
    use_bias: bool = True,
    dropout: Optional[float] = None,
    normalization: Optional[Union[str, tf.keras.layers.Layer]] = "batch_norm",
    depth: int = 1,
    **dense_kwargs,
) -> Block:
    """A block that applies a dense residual block to the input.
    The residual consists in the input summed element-wise with the
    output of the dense block.
    The dense block projects the inputs using dense layers to the same output dim.
    If the input dim is very high dimensional, the low_rank_dim can
    be used to create a low rank matrix and reduce the necessary number
    of parameters for this projection.

    Example usage::
        block = ml.DenseResidualBlock(depth=3).connect(ml.MLPBlock([1]))

    Parameters
    ----------
    low_rank_dim: int, optional
        The dimension of the low rank matrix. If set, it projects the input to the
        low_rank_dim (`LR`) and then back to the input dim (`I`) as output.
        That requires much less parameters (`I*LR + LR*I`) than projecting the
        inputs dim directly to the same dim (`I*I`). By default None
    activation: Union[str, tf.keras.layers.Layer], optional
        The activation function to use. By default "relu"
    use_bias: bool
        Whether to use a bias in the MLP. By default True
    dropout: float
        The dropout rate to use. By default 0.0
    normalization: Union[str, tf.keras.layers.Layer], optional
        The normalization layer to use. By the default None.
    depth: int
        The number of residual blocks to stack. By default 1
    """

    block_layers = []
    block_layers.append(
        DenseMaybeLowRank(low_rank_dim, activation=None, use_bias=use_bias, **dense_kwargs)
    )
    if dropout:
        block_layers.append(tf.keras.layers.Dropout(dropout))
    if normalization:
        if normalization == "batch_norm":
            block_layers.append(tf.keras.layers.BatchNormalization())
        elif isinstance(normalization, tf.keras.layers.Layer):
            block_layers.append(normalization)
        else:
            raise ValueError("Normalization needs to be an instance `Layer` or " "`batch_norm`")

    output = ResidualBlock(
        SequentialBlock(block_layers, block_name="DenseResidual"), activation=activation
    )

    if depth > 1:
        return output.repeat(depth - 1)
    elif depth < 1:
        raise ValueError(
            "The depth (number of stacked residual blocks) needs " "to be equal or greater than 1."
        )

    return output


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class _Dense(tf.keras.layers.Layer):
    def __init__(
        self,
        units,
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        pre_aggregation="concat",
        dense=None,
        **kwargs,
    ):
        super(_Dense, self).__init__(**kwargs)
        self.dense = dense or tf.keras.layers.Dense(
            units,
            activation,
            use_bias,
            kernel_initializer,
            bias_initializer,
            kernel_regularizer,
            bias_regularizer,
            activity_regularizer,
            kernel_constraint,
            bias_constraint,
            **kwargs,
        )
        self.pre_aggregation = pre_aggregation
        self.units = units

    def call(self, inputs, **kwargs):
        if isinstance(inputs, dict):
            inputs = tabular_aggregation_registry.parse(self.pre_aggregation)(inputs)

        filtered_kwargs = filter_kwargs(kwargs, self.dense)
        return self.dense(inputs, **filtered_kwargs)

    def compute_output_shape(self, input_shape):
        if isinstance(input_shape, dict):
            agg = tabular_aggregation_registry.parse(self.pre_aggregation)
            input_shape = agg.compute_output_shape(input_shape)

        return self.dense.compute_output_shape(input_shape)

    def get_config(self):
        config = super(_Dense, self).get_config()
        config["pre_aggregation"] = self.pre_aggregation
        config["units"] = self.units

        return maybe_serialize_keras_objects(self, config, ["dense"])

    @classmethod
    def from_config(cls, config):
        config = maybe_deserialize_keras_objects(config, {"dense": tf.keras.layers.deserialize})

        return cls(**config)


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class DenseMaybeLowRank(tf.keras.layers.Layer):
    def __init__(
        self,
        low_rank_dim: Optional[int] = None,
        use_bias: bool = True,
        activation: Optional[Union[str, tf.keras.layers.Layer]] = None,
        kernel_initializer: InitializerType = "truncated_normal",
        bias_initializer: InitializerType = "zeros",
        kernel_regularizer: Optional[RegularizerType] = None,
        bias_regularizer: Optional[RegularizerType] = None,
        pre_aggregation: Optional[TabularAggregationType] = "concat",
        dense: Optional[tf.keras.layers.Dense] = None,
        dense_u: Optional[tf.keras.layers.Dense] = None,
        **kwargs,
    ):
        """A block that projects the inputs the same input dim.
        If the input dim is very high dimensional, the low_rank_dim can
        be used to create a low rank matrix and reduce the necessary number
        of parameters for this projection.

        Parameters
        ----------
        low_rank_dim : Optional[int], optional
            The dimension of the low rank matrix. If set, it projects the input to the
            low_rank_dim (`LR`) and then back to the input dim (`I`) as output.
            That requires much less parameters (`I*LR + LR*I`) than projecting the
            inputs dim directly to the same dim (`I*I`), by default None.
        use_bias : bool, optional
            Whether to use a bias in the MLP, by default True
        activation : Optional[Union[str,tf.keras.layers.Layer]], optional
            The activation function to use. By default None
        kernel_initializer: InitializerType
            Initializer for the kernel weights matrix. Defaults to "glorot_uniform".
        bias_initializer: InitializerType
            Initializer for the bias vector. Default to "zeros".
        kernel_regularizer: Optional[RegularizerType]
            Regularizer function applied to the kernel weights matrix. Default to None.
        bias_regularizer: Optional[RegularizerType]
            Regularizer function applied to the bias vector.  Default to None.
        pre_aggregation : Optional[TabularAggregationType], optional
            Aggregation to be done before projection, by default "concat"
        dense : Optional[tf.keras.layers.Dense], optional
            An optional dense layer to be used for projection,
            by default None. If not set it is created internally.
        dense_u : Optional[tf.keras.layers.Dense], optional
            An optional dense layer to be called first if low_rank_dim is set,
            by default None. If not set it is created internally.
        """

        super().__init__(**kwargs)
        self.low_rank_dim = low_rank_dim
        self.use_bias = use_bias
        self.activation = activation
        self.kernel_initializer = kernel_initializer
        self.bias_initializer = bias_initializer
        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer
        self.pre_aggregation = pre_aggregation
        self.dense = dense
        self.dense_u = dense_u

    def build(self, input_shape):
        last_dim = input_shape[-1]

        if self.dense is None:
            self.dense = _Dense(
                last_dim,
                activation=self.activation,
                kernel_initializer=self.kernel_initializer,
                bias_initializer=self.bias_initializer,
                kernel_regularizer=self.kernel_regularizer,
                bias_regularizer=self.bias_regularizer,
                use_bias=self.use_bias,
            )

        if self.low_rank_dim is not None and self.dense_u is None:
            self.dense_u = _Dense(
                self.low_rank_dim,
                activation=self.activation,
                kernel_initializer=self.kernel_initializer,
                kernel_regularizer=self.kernel_regularizer,
                use_bias=False,
            )
        super(DenseMaybeLowRank, self).build(input_shape)

    def call(self, inputs: tf.Tensor, **kwargs) -> tf.Tensor:
        if isinstance(inputs, dict):
            inputs = tabular_aggregation_registry.parse(self.pre_aggregation)(inputs)

        if self.low_rank_dim is None:
            return self.dense(inputs)  # type: ignore

        return self.dense(self.dense_u(inputs))  # type: ignore

    def compute_output_shape(self, input_shape):
        if isinstance(input_shape, dict):
            agg = tabular_aggregation_registry.parse(self.pre_aggregation)
            input_shape = agg.compute_output_shape(input_shape)

        return input_shape

    def get_config(self):
        config = dict(
            low_rank_dim=self.low_rank_dim,
            use_bias=self.use_bias,
            activation=self.activation,
            pre_aggregation=self.pre_aggregation,
        )
        config.update(super(DenseMaybeLowRank, self).get_config())

        config = maybe_serialize_keras_objects(
            self,
            config,
            [
                "dense",
                "dense_u",
            ],
        )

        return config

    @classmethod
    def from_config(cls, config):
        config = maybe_deserialize_keras_objects(
            config,
            [
                "dense",
                "dense_u",
            ],
        )

        return cls(**config)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from merlin.schema import ColumnSchema, Tags


def _augment_schema(
    schema,
    cats=None,
    conts=None,
    labels=None,
    sparse_names=None,
    sparse_max=None,
    sparse_as_dense=False,
):
    labels = [labels] if isinstance(labels, str) else labels
    for label in labels or []:
        schema[label] = schema[label].with_tags(Tags.TARGET)
    for label in cats or []:
        schema[label] = schema[label].with_tags(Tags.CATEGORICAL)
    for label in conts or []:
        schema[label] = schema[label].with_tags(Tags.CONTINUOUS)

    # Set the appropriate properties for the sparse_names/sparse_max/sparse_as_dense
    for col in sparse_names or []:
        cs = schema[col]
        properties = cs.properties
        if sparse_max and col in sparse_max:
            properties["value_count"] = {"max": sparse_max[col]}
        schema[col] = ColumnSchema(
            name=cs.name,
            tags=cs.tags,
            dtype=cs.dtype,
            is_list=True,
            is_ragged=not sparse_as_dense,
            properties=properties,
        )

    return schema



# Test is currently breaks in TF 2.10

from testbook import testbook

from tests.conftest import REPO_ROOT

p = "examples/usecases/incremental-training-with-layer-freezing.ipynb"


@testbook(
    REPO_ROOT / p,
    timeout=180,
    execute=False,
)
def test_usecase_incremental_training_layer_freezing(tb):
    tb.inject(
        """
        import os
        os.environ["NUM_ROWS"] = "1000"
        """
    )
    tb.execute()
    model = tb.ref("model")
    assert set(model.history.history.keys()) == set(
        [
            "loss",
            "loss_batch",
            "recall_at_10",
            "mrr_at_10",
            "ndcg_at_10",
            "map_at_10",
            "precision_at_10",
            "regularization_loss",
        ]
    )



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from types import SimpleNamespace
from typing import Dict

import tensorflow as tf
from tensorflow.keras.layers import Layer
from transformers.modeling_tf_outputs import TFBaseModelOutputWithPoolingAndCrossAttentions
from transformers.modeling_tf_utils import TFSequenceSummary

from merlin.models.tf.core.base import Block


@Block.registry.register("last_hidden_state")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class LastHiddenState(Layer):
    """Select the hidden states of the last transform layer
    from the HF dataclass object
    `TFBaseModelOutputWithPoolingAndCrossAttentions`

    Parameters
    ----------
    inputs: TFBaseModelOutputWithPoolingAndCrossAttentions
        The output class returned by the HuggingFace transformer layer
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.supports_masking = True

    def call(self, inputs: TFBaseModelOutputWithPoolingAndCrossAttentions):
        return inputs.last_hidden_state


@Block.registry.register("inference_hidden_state")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class TransformerInferenceHiddenState(Layer):
    """A post-processing layer to select the hidden state
    of the next-item position, during inference.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.supports_masking = True

    def call(
        self,
        inputs: tf.Tensor,
        training: bool = False,
        testing: bool = False,
    ):
        """Select the hidden state of the target (last) position, during inference.
        During training or testing, the inputs are returned
        without any processing.

        Parameters:
        ----------
        inputs: tf.Tensor
            The 3-D output tensor returned by the transformer block
        training : bool, optional
            Flag that indicates whether in training mode, by default True
        testing : bool, optional
            Flag that indicates whether in evaluation mode, by default True

        Returns
        -------
        tf.Tensor
            If inference, returns a 2-D tensor with the hidden states of
            the target position
        """
        batch_size = tf.shape(inputs)[0]
        if not training and not testing:
            if getattr(inputs, "_keras_mask", None) is not None:
                inputs = tf.reshape(
                    tf.boolean_mask(inputs, inputs._keras_mask), (-1, inputs.shape[-1])
                )
        tf.debugging.assert_equal(
            tf.shape(inputs)[0],
            batch_size,
            f"The resulting tensor has {tf.shape(inputs)[0]} rows, which does not match"
            f" the inputs batch-size {batch_size}. During inference only one position "
            "candidate (the last one) should be masked per example",
        )
        return inputs


@Block.registry.register("pooler_output")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class PoolerOutput(Layer):
    """Select the pooled representation of the sequence
    from the HF dataclass object
    `TFBaseModelOutputWithPoolingAndCrossAttentions`

    Parameters
    ----------
    inputs: TFBaseModelOutputWithPoolingAndCrossAttentions
        The output class returned by the HuggingFace transformer layer
    """

    def call(self, inputs: TFBaseModelOutputWithPoolingAndCrossAttentions):
        return inputs.pooler_output


@Block.registry.register("hidden_states")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class HiddenStates(Layer):
    """Select the tuple of hidden states of all transformer layers
    from the HF dataclass object
    `TFBaseModelOutputWithPoolingAndCrossAttentions`

    Parameters
    ----------
    inputs: TFBaseModelOutputWithPoolingAndCrossAttentions
        The output class returned by the HuggingFace transformer layer
    """

    def call(self, inputs: TFBaseModelOutputWithPoolingAndCrossAttentions):
        return inputs.hidden_states


@Block.registry.register("attentions")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class AttentionWeights(Layer):
    """Select the attention weights of all transformer layers
    from the HF dataclass object
    `TFBaseModelOutputWithPoolingAndCrossAttentions`

    Parameters
    ----------
    inputs: TFBaseModelOutputWithPoolingAndCrossAttentions
        The output class returned by the HuggingFace transformer layer
    """

    def call(self, inputs: TFBaseModelOutputWithPoolingAndCrossAttentions):
        return inputs.attentions


@Block.registry.register("last_hidden_state_and_attention")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class LastHiddenStateAndAttention(Layer):
    """Select the hidden states and attention weights of the last
    tranfrom the HF dataclass object
    `TFBaseModelOutputWithPoolingAndCrossAttentions`

    Parameters
    ----------
    inputs: TFBaseModelOutputWithPoolingAndCrossAttentions
        The output class returned by the HuggingFace transformer layer
    """

    def call(self, inputs: TFBaseModelOutputWithPoolingAndCrossAttentions):
        return (inputs.last_hidden_state, inputs.attentions[-1])


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class PrepareTransformerInputs(tf.keras.layers.Layer):
    """Prepare the dictionary of inputs expected by the transformer layer"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.supports_masking = True

    def call(self, inputs: tf.Tensor) -> Dict[str, tf.Tensor]:
        mask = None
        if getattr(inputs, "_keras_mask", None) is not None and isinstance(
            inputs._keras_mask, tf.RaggedTensor
        ):
            mask = inputs._keras_mask.to_tensor()
        if isinstance(inputs, tf.RaggedTensor):
            # convert to a dense tensor as HF transformers do not support ragged tensors
            inputs = inputs.to_tensor()
        if mask is not None:
            inputs._keras_mask = mask
        return {"inputs_embeds": inputs}


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class SequenceSummary(TFSequenceSummary):
    def __init__(self, summary: str, initializer_range: float = 0.02, **kwargs):
        self.summary = summary
        config = SimpleNamespace(summary_type=summary)
        super().__init__(config, initializer_range=initializer_range, **kwargs)

    def get_config(self):
        config = super().get_config()
        config["summary"] = self.summary
        return config

    @classmethod
    def from_config(cls, config, custom_objects=None):
        output = SequenceSummary(**config)
        output.__class__ = cls
        return output


@Block.registry.register("sequence_last")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class SequenceLast(SequenceSummary):
    def __init__(self, initializer_range: float = 0.02, **kwargs):
        super().__init__("last", initializer_range=initializer_range, **kwargs)


@Block.registry.register("sequence_first")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class SequenceFirst(SequenceSummary):
    def __init__(self, initializer_range: float = 0.02, **kwargs):
        super().__init__("first", initializer_range=initializer_range, **kwargs)


@Block.registry.register("sequence_mean")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class SequenceMean(SequenceSummary):
    def __init__(self, initializer_range: float = 0.02, **kwargs):
        super().__init__("mean", initializer_range=initializer_range, **kwargs)


@Block.registry.register("sequence_cls_index")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class SequenceClsIndex(SequenceSummary):
    def __init__(self, initializer_range: float = 0.02, **kwargs):
        super().__init__("cls_index", initializer_range=initializer_range, **kwargs)



from typing import Any, Callable, Dict, List, Optional, Sequence, Union

import tensorflow as tf

from merlin.models.tf.blocks.mlp import MLPBlock
from merlin.models.tf.blocks.retrieval.matrix_factorization import QueryItemIdsEmbeddingsBlock
from merlin.models.tf.blocks.retrieval.two_tower import TwoTowerBlock
from merlin.models.tf.blocks.sampling.base import ItemSampler
from merlin.models.tf.core.base import Block, BlockType
from merlin.models.tf.core.combinators import ParallelBlock
from merlin.models.tf.core.encoder import EmbeddingEncoder, Encoder
from merlin.models.tf.inputs.base import InputBlock
from merlin.models.tf.inputs.embedding import EmbeddingOptions
from merlin.models.tf.models.base import Model, RetrievalModel, RetrievalModelV2
from merlin.models.tf.models.utils import parse_prediction_tasks
from merlin.models.tf.outputs.base import DotProduct, ModelOutput
from merlin.models.tf.outputs.contrastive import ContrastiveOutput
from merlin.models.tf.outputs.sampling.base import ItemSamplersType
from merlin.models.tf.outputs.sampling.popularity import PopularityBasedSamplerV2
from merlin.models.tf.prediction_tasks.base import ParallelPredictionBlock, PredictionTask
from merlin.models.tf.prediction_tasks.next_item import NextItemPredictionTask
from merlin.models.tf.prediction_tasks.retrieval import ItemRetrievalTask
from merlin.models.utils.schema_utils import categorical_cardinalities
from merlin.schema import Schema, Tags


def MatrixFactorizationModel(
    schema: Schema,
    dim: int,
    query_id_tag=Tags.USER_ID,
    item_id_tag=Tags.ITEM_ID,
    embeddings_initializers: Optional[
        Union[Dict[str, Callable[[Any], None]], Callable[[Any], None]]
    ] = None,
    embeddings_l2_reg: float = 0.0,
    post: Optional[BlockType] = None,
    prediction_tasks: Optional[
        Union[PredictionTask, List[PredictionTask], ParallelPredictionBlock]
    ] = None,
    logits_temperature: float = 1.0,
    samplers: Sequence[ItemSampler] = (),
    **kwargs,
) -> RetrievalModel:
    """Builds a matrix factorization model.

    Example Usage::
        mf = MatrixFactorizationModel(schema, dim=128)
        mf.compile(optimizer="adam")
        mf.fit(train_data, epochs=10)

    Parameters
    ----------
    schema: Schema
        The `Schema` with the input features
    dim: int
        The dimension of the embeddings.
    query_id_tag : Tag
        The tag to select query features, by default `Tags.USER`
    item_id_tag : Tag
        The tag to select item features, by default `Tags.ITEM`
    embeddings_initializers : Optional[Dict[str, Callable[[Any], None]]] = None
        An initializer function or a dict where keys are feature names and values are
        callable to initialize embedding tables
    embeddings_l2_reg: float = 0.0
        Factor for L2 regularization of the embeddings vectors (from the current batch only)
    post: Optional[Block], optional
        The optional `Block` to apply on both outputs of Two-tower model
    prediction_tasks: optional
        The optional `PredictionTask` or list of `PredictionTask` to apply on the model.
    logits_temperature: float
        Parameter used to reduce model overconfidence, so that logits / T.
        Defaults to 1.
    loss: Optional[LossType]
        Loss function.
        Defaults to `bpr`.
    samplers: List[ItemSampler]
        List of samplers for negative sampling, by default `[InBatchSampler()]`

    Returns
    -------
    RetrievalModel
    """

    if not prediction_tasks:
        prediction_tasks = ItemRetrievalTask(
            schema,
            logits_temperature=logits_temperature,
            samplers=list(samplers),
            **kwargs,
        )

    prediction_tasks = parse_prediction_tasks(schema, prediction_tasks)
    mf = QueryItemIdsEmbeddingsBlock(
        schema=schema,
        dim=dim,
        query_id_tag=query_id_tag,
        item_id_tag=item_id_tag,
        embeddings_initializers=embeddings_initializers,
        embeddings_l2_reg=embeddings_l2_reg,
        post=post,
        **kwargs,
    )

    model = RetrievalModel(mf, prediction_tasks)

    return model


def TwoTowerModel(
    schema: Schema,
    query_tower: Block,
    item_tower: Optional[Block] = None,
    query_tower_tag=Tags.USER,
    item_tower_tag=Tags.ITEM,
    embedding_options: EmbeddingOptions = EmbeddingOptions(
        embedding_dims=None,
        embedding_dim_default=64,
        infer_embedding_sizes=False,
        infer_embedding_sizes_multiplier=2.0,
    ),
    post: Optional[BlockType] = None,
    prediction_tasks: Optional[
        Union[PredictionTask, List[PredictionTask], ParallelPredictionBlock]
    ] = None,
    logits_temperature: float = 1.0,
    samplers: Sequence[ItemSampler] = (),
    **kwargs,
) -> RetrievalModel:
    """Builds the Two-tower architecture, as proposed in [1].

    Example Usage::
        two_tower = TwoTowerModel(schema, MLPBlock([256, 64]))
        two_tower.compile(optimizer="adam")
        two_tower.fit(train_data, epochs=10)

    References
    ----------
    [1] Yi, Xinyang, et al.
        "Sampling-bias-corrected neural modeling for large corpus item recommendations."
        Proceedings of the 13th ACM Conference on Recommender Systems. 2019.

    Parameters
    ----------
    schema: Schema
        The `Schema` with the input features
    query_tower: Block
        The `Block` that combines user features
    item_tower: Optional[Block], optional
        The optional `Block` that combines items features.
        If not provided, a copy of the query_tower is used.
    query_tower_tag: Tag
        The tag to select query features, by default `Tags.USER`
    item_tower_tag: Tag
        The tag to select item features, by default `Tags.ITEM`
    embedding_options : EmbeddingOptions
        Options for the input embeddings.
        - embedding_dims: Optional[Dict[str, int]] - The dimension of the
        embedding table for each feature (key), by default {}
        - embedding_dim_default: int - Default dimension of the embedding
        table, when the feature is not found in ``embedding_dims``, by default 64
        - infer_embedding_sizes : bool, Automatically defines the embedding
        dimension from the feature cardinality in the schema, by default False
        - infer_embedding_sizes_multiplier: int. Multiplier used by the heuristic
        to infer the embedding dimension from its cardinality. Generally
        reasonable values range between 2.0 and 10.0. By default 2.0.
    post: Optional[Block], optional
        The optional `Block` to apply on both outputs of Two-tower model
    prediction_tasks: optional
        The optional `PredictionTask` or list of `PredictionTask` to apply on the model.
    logits_temperature: float
        Parameter used to reduce model overconfidence, so that logits / T.
        Defaults to 1.
    loss: Optional[LossType]
        Loss function.
        Defaults to `categorical_crossentropy`.
    samplers: List[ItemSampler]
        List of samplers for negative sampling, by default `[InBatchSampler()]`

    Returns
    -------
    RetrievalModel
    """

    if not prediction_tasks:
        prediction_tasks = ItemRetrievalTask(
            schema,
            logits_temperature=logits_temperature,
            samplers=list(samplers),
            **kwargs,
        )

    prediction_tasks = parse_prediction_tasks(schema, prediction_tasks)
    two_tower = TwoTowerBlock(
        schema=schema,
        query_tower=query_tower,
        item_tower=item_tower,
        query_tower_tag=query_tower_tag,
        item_tower_tag=item_tower_tag,
        embedding_options=embedding_options,
        post=post,
        **kwargs,
    )

    model = RetrievalModel(two_tower, prediction_tasks)

    return model


def YoutubeDNNRetrievalModel(
    schema: Schema,
    aggregation: str = "concat",
    top_block: Block = MLPBlock([64]),
    l2_normalization: bool = True,
    extra_pre_call: Optional[Block] = None,
    task_block: Optional[Block] = None,
    logits_temperature: float = 1.0,
    sampled_softmax: bool = True,
    num_sampled: int = 100,
    min_sampled_id: int = 0,
    embedding_options: EmbeddingOptions = EmbeddingOptions(
        embedding_dims=None,
        embedding_dim_default=64,
        infer_embedding_sizes=False,
        infer_embedding_sizes_multiplier=2.0,
    ),
) -> Model:
    """Build the Youtube-DNN retrieval model.
    More details of the architecture can be found in [1]_.
    The sampled_softmax is enabled by default [2]_ [3]_ [4]_.

    Example Usage::
        model = YoutubeDNNRetrievalModel(schema, num_sampled=100)
        model.compile(optimizer="adam")
        model.fit(train_data, epochs=10)

    References
    ----------
    .. [1] Covington, Paul, Jay Adams, and Emre Sargin.
        "Deep neural networks for youtube recommendations."
        Proceedings of the 10th ACM conference on recommender systems. 2016.

    .. [2] Yoshua Bengio and Jean-Sébastien Sénécal. 2003. Quick Training of Probabilistic
       Neural Nets by Importance Sampling. In Proceedings of the conference on Artificial
       Intelligence and Statistics (AISTATS).

    .. [3] Y. Bengio and J. S. Senecal. 2008. Adaptive Importance Sampling to Accelerate
       Training of a Neural Probabilistic Language Model. Trans. Neur. Netw. 19, 4 (April
       2008), 713–722. https://doi.org/10.1109/TNN.2007.912312

    .. [4] Jean, Sébastien, et al. "On using very large target vocabulary for neural
        machine translation." arXiv preprint arXiv:1412.2007 (2014).


    Parameters
    ----------
    schema: Schema
        The `Schema` with the input features
    aggregation: str
        The aggregation method to use for the sequence of features.
        Defaults to `concat`.
    top_block: Block
        The `Block` that combines the top features
    l2_normalization: bool
        Whether to apply L2 normalization before computing dot interactions.
        Defaults to True.
    extra_pre_call: Optional[Block]
        The optional `Block` to apply before the model.
    task_block: Optional[Block]
        The optional `Block` to apply on the model.
    logits_temperature: float
        Parameter used to reduce model overconfidence, so that logits / T.
        Defaults to 1.
    sampled_softmax: bool
        Compute the logits scores over all items of the catalog or
        generate a subset of candidates
        Defaults to False
    num_sampled: int
        When sampled_softmax is enabled, specify the number of
        negative candidates to generate for each batch.
        Defaults to 100
    min_sampled_id: int
        The minimum id value to be sampled with sampled softmax.
        Useful to ignore the first categorical
        encoded ids, which are usually reserved for <nulls>,
        out-of-vocabulary or padding. Defaults to 0.
    embedding_options : EmbeddingOptions, optional
        An EmbeddingOptions instance, which allows for a number of
        options for the embedding table, by default EmbeddingOptions()
    """

    inputs = InputBlock(
        schema,
        aggregation=aggregation,
        embedding_options=embedding_options,
    )

    task = NextItemPredictionTask(
        schema=schema,
        weight_tying=False,
        extra_pre_call=extra_pre_call,
        task_block=task_block,
        logits_temperature=logits_temperature,
        l2_normalization=l2_normalization,
        sampled_softmax=sampled_softmax,
        num_sampled=num_sampled,
        min_sampled_id=min_sampled_id,
    )

    # TODO: Figure out how to make this fit as
    # a RetrievalModel (which must have a RetrievalBlock)
    return Model(inputs, top_block, task)


def MatrixFactorizationModelV2(
    schema: Schema,
    dim: int,
    query_id_tag=Tags.USER_ID,
    candidate_id_tag=Tags.ITEM_ID,
    embeddings_initializers: Optional[
        Union[Dict[str, Callable[[Any], None]], Callable[[Any], None]]
    ] = None,
    embeddings_l2_batch_regularization: Union[float, Dict[str, float]] = 0.0,
    post: Optional[BlockType] = None,
    outputs: Optional[Union[ModelOutput, List[ModelOutput]]] = None,
    negative_samplers: ItemSamplersType = None,
    logits_temperature: float = 1.0,
    **kwargs,
) -> RetrievalModelV2:
    """Builds a matrix factorization (MF) model.

    Example Usage::
        mf = MatrixFactorizationModelV2(schema, dim=128)
        mf.compile(optimizer="adam")
        mf.fit(train_data, epochs=10)

    Parameters
    ----------
    schema: Schema
        The `Schema` with the input features
    dim: int
        The dimension of the embeddings.
    query_id_tag : Tag
        The tag to select query-id feature, by default `Tags.USER_ID`
    candidate_id_tag : Tag
        The tag to select cnadidate-id feature, by default `Tags.ITEM_ID`
    embeddings_initializers : Optional[Dict[str, Callable[[Any], None]]] = None
        An initializer function or a dict where keys are feature names and values are
        callable to initialize embedding tables
    embeddings_l2_batch_regularization: Union[float, Dict[str, float]] = 0.0
        Factor for L2 regularization of the embeddings vectors (from the current batch only)
        If a dictionary is provided, the keys are the query-id and candidate-id names and
        the values are the regularization factor.
    post: Optional[Block], optional
        The optional `Block` to apply on both outputs of the MF towers.
    outputs: optional
        The optional `ModelOutput` or list of `ModelOutput` to apply on the MF model.
    negative_samplers: List[ItemSampler]
        List of samplers for negative sampling, by default None
        If the `outputs` and `negative_samplers` are not specified the Matrix Factorization model
        is trained with contrastive learning and `in-batch` negative sampling strategy.
    logits_temperature: float
        Parameter used to reduce model overconfidence, so that logits / T.
        Defaults to 1.


    Returns
    -------
    RetrievalModelV2
    """

    query = schema.select_by_tag(query_id_tag)
    candidate = schema.select_by_tag(candidate_id_tag)

    query_encoder = EmbeddingEncoder(
        query,
        dim=dim,
        embeddings_initializer=embeddings_initializers,
        embeddings_l2_batch_regularization=embeddings_l2_batch_regularization,
        post=post,
    )
    candidate_encoder = EmbeddingEncoder(
        candidate,
        dim=dim,
        embeddings_initializer=embeddings_initializers,
        embeddings_l2_batch_regularization=embeddings_l2_batch_regularization,
        post=post,
    )

    if not outputs:
        if not negative_samplers:
            negative_samplers = ["in-batch"]
        outputs = ContrastiveOutput(
            to_call=DotProduct(),
            negative_samplers=negative_samplers,
            logits_temperature=logits_temperature,
            schema=candidate,
            **kwargs,
        )

    if isinstance(outputs, list):
        outputs = ParallelBlock(*outputs)

    model = RetrievalModelV2(
        query=query_encoder,
        candidate=candidate_encoder,
        output=outputs,
    )

    return model


def TwoTowerModelV2(
    query_tower: Encoder,
    candidate_tower: Encoder,
    candidate_id_tag=Tags.ITEM_ID,
    outputs: Optional[Union[ModelOutput, List[ModelOutput]]] = None,
    logits_temperature: float = 1.0,
    negative_samplers: ItemSamplersType = None,
    **kwargs,
) -> RetrievalModelV2:
    """Builds the Two-tower architecture, as proposed in [1].

    Example Usage::
        query = mm.Encoder(user_schema, mm.MLPBlock([128]))
        candidate = mm.Encoder(item_schema, mm.MLPBlock([128]))
        model = TwoTowerModel(query, candidate)
        two_tower.compile(optimizer="adam")
        two_tower.fit(train_data, epochs=10)

    References
    ----------
    [1] Yi, Xinyang, et al.
        "Sampling-bias-corrected neural modeling for large corpus item recommendations."
        Proceedings of the 13th ACM Conference on Recommender Systems. 2019.

    Parameters
    ----------
    query_tower: Encoder
        The layer that encodes query features
    candidate_tower: Encoder
        The  layer that encodes candidates features
    candidate_id_tag: Tag, optional
        The tag to select candidate-id feature, by default `Tags.ITEM_ID`
    outputs:  Union[ModelOutput, List[ModelOutput]], optional
        The optional `ModelOutput` or list of `ModelOutput` to apply on the model.
    logits_temperature: float
        Parameter used to reduce model overconfidence, so that logits / T.
        Defaults to 1.
    negative_samplers: List[ItemSampler]
        List of samplers for negative sampling, by default None
        If the `outputs` and `negative_samplers` are not specified the two tower model
        is trained with contrastive learning and `in-batch` negative sampling strategy.

    Returns
    -------
    RetrievalModelV2
    """
    assert isinstance(query_tower, Encoder), ValueError(
        "The query tower should be an instance of `Encoder` class"
    )
    assert isinstance(candidate_tower, Encoder), ValueError(
        "The query tower should be an instance of `Encoder` class"
    )

    if not outputs:
        if not negative_samplers:
            negative_samplers = ["in-batch"]
        outputs = ContrastiveOutput(
            to_call=DotProduct(),
            negative_samplers=negative_samplers,
            logits_temperature=logits_temperature,
            schema=candidate_tower._schema.select_by_tag(candidate_id_tag),
            **kwargs,
        )

    if isinstance(outputs, list):
        outputs = ParallelBlock(*outputs)

    model = RetrievalModelV2(
        query=query_tower,
        candidate=candidate_tower,
        output=outputs,
    )

    return model


def YoutubeDNNRetrievalModelV2(
    schema: Schema,
    candidate_id_tag=Tags.ITEM_ID,
    top_block: Optional[tf.keras.layers.Layer] = MLPBlock([64]),
    post: Optional[tf.keras.layers.Layer] = None,
    inputs: tf.keras.layers.Layer = None,
    outputs: Optional[Union[ModelOutput, List[ModelOutput]]] = None,
    logits_temperature: float = 1.0,
    num_sampled: int = 100,
    min_sampled_id: int = 0,
    **kwargs,
) -> RetrievalModelV2:
    """Build the Youtube-DNN retrieval model.
    More details of the architecture can be found in [1]_.
    Training with sampled_softmax is enabled by default [2]_ [3]_ [4]_.

    Example Usage::
        model = YoutubeDNNRetrievalModelV2(schema, num_sampled=100)
        model.compile(optimizer="adam")
        model.fit(train_data, epochs=10)

    References
    ----------
    .. [1] Covington, Paul, Jay Adams, and Emre Sargin.
        "Deep neural networks for youtube recommendations."
        Proceedings of the 10th ACM conference on recommender systems. 2016.

    .. [2] Yoshua Bengio and Jean-Sébastien Sénécal. 2003. Quick Training of Probabilistic
       Neural Nets by Importance Sampling. In Proceedings of the conference on Artificial
       Intelligence and Statistics (AISTATS).

    .. [3] Y. Bengio and J. S. Senecal. 2008. Adaptive Importance Sampling to Accelerate
       Training of a Neural Probabilistic Language Model. Trans. Neur. Netw. 19, 4 (April
       2008), 713–722. https://doi.org/10.1109/TNN.2007.912312

    .. [4] Jean, Sébastien, et al. "On using very large target vocabulary for neural
        machine translation." arXiv preprint arXiv:1412.2007 (2014).


    Parameters
    ----------
    schema: Schema
        The `Schema` with the input features
    candidate_id_tag : Tag
        The tag to select cnadidate-id feature, by default `Tags.ITEM_ID`
    top_block: tf.keras.layers.Layer
        The hidden layers to apply on top of the features representation
        vector.
    inputs: tf.keras.layers.Layer, optional
        The input layer to encode input features (sparse and context features)
        If not specified, the input layer is inferred from the schema
        By default None
    post: Optional[tf.keras.layers.Layer], optional
        The optional layer to apply on top of the query encoder.
    logits_temperature: float, optional
        Parameter used to reduce model overconfidence, so that logits / T.
        Defaults to 1.
    num_sampled: int, optional
        When sampled_softmax is enabled, specify the number of
        negative candidates to generate for each batch.
        By default 100
    min_sampled_id: int, optional
        The minimum id value to be sampled with sampled softmax.
        Useful to ignore the first categorical
        encoded ids, which are usually reserved for <nulls>,
        out-of-vocabulary or padding.
        By default 0.
    """
    if not inputs:
        inputs = schema

    candidate = schema.select_by_tag(candidate_id_tag)
    if not candidate:
        raise ValueError(f"The schema should contain a feature tagged as `{candidate_id_tag}`")

    query = Encoder(inputs, top_block, post=post)

    if not outputs:
        cardinalities = categorical_cardinalities(candidate)
        candidate = candidate.first
        candidate_table = query.first["categorical"][candidate.name]
        num_classes = cardinalities[candidate.name]

        outputs = ContrastiveOutput(
            to_call=candidate_table,
            logits_temperature=logits_temperature,
            post=post,
            negative_samplers=PopularityBasedSamplerV2(
                max_num_samples=num_sampled, max_id=num_classes - 1, min_id=min_sampled_id
            ),
        )

    return RetrievalModelV2(query=query, output=outputs)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from merlin.datasets.ecommerce.aliccp.dataset import (
    default_aliccp_transformation,
    get_aliccp,
    prepare_alliccp,
    transform_aliccp,
)
from merlin.datasets.ecommerce.booking.dataset import get_booking, transform_booking
from merlin.datasets.ecommerce.dressipi.dataset import get_dressipi2022

__all__ = [
    "prepare_alliccp",
    "transform_aliccp",
    "get_aliccp",
    "default_aliccp_transformation",
    "get_dressipi2022",
    "get_booking",
    "transform_booking",
]



from testbook import testbook

from tests.conftest import REPO_ROOT


@testbook(
    REPO_ROOT / "examples/usecases/entertainment-with-pretrained-embeddings.ipynb", execute=False
)
def test_usecase_pretrained_embeddings(tb):
    tb.execute()
    history = tb.ref("history")
    assert set(history.keys()) == set(["auc", "loss", "loss_batch", "regularization_loss"])



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Protocol, runtime_checkable


@runtime_checkable
class MerlinModel(Protocol):
    """Protocol for a merlin model.

    This defines common methods that should be implemented on all model implementations.
    Including those with different frameworks (e.g. Tensorflow, XGBoost, Pytorch)
    """

    def save(self, path) -> None:
        """Save the model to a local path.

        Parameters
        ----------
        path : Union[str, os.PathLike]
        """
        ...

    @classmethod
    def load(cls, path):
        """Load the model from a path-like argument provided where a model was previously saved.

        Parameters
        ----------
        path : Union[str, os.PathLike]
            A path correspoonding to the directory where a model was previously saved.
        """
        ...

    def fit(self, dataset, *args, **kwargs):
        """Fit the model on the provided dataset.

        Parameters
        ----------
        dataset : merlin.io.Dataset
            The training dataset to be used to fit the model.
        """
        ...

    def evaluate(self, dataset, *args, **kwargs):
        """Return evaluation metrics on a dataset.

        Parameters
        ----------
        dataset : merlin.io.Dataset
            The evaluation dataset to be used to compute metrics.
        """
        ...

    def predict(self, dataset, *args, **kwargs):
        """Return predictions generated by the model.

        Parameters
        ----------
        dataset : merlin.io.Dataset
            The dataset to generate predictions from.
        """
        ...



import itertools

import numpy as np
import pytest
import tensorflow as tf
from transformers import BertConfig

import merlin.models.tf as mm
from merlin.io import Dataset
from merlin.models.tf.loader import Loader
from merlin.models.tf.transformers.block import (
    AlbertBlock,
    BertBlock,
    GPT2Block,
    RobertaBlock,
    XLNetBlock,
)
from merlin.models.tf.utils import testing_utils
from merlin.schema import Tags


def test_import():
    import transformers

    assert transformers is not None


@pytest.mark.parametrize("run_eagerly", [True])
def test_retrieval_transformer(sequence_testing_data: Dataset, run_eagerly):

    sequence_testing_data.schema = sequence_testing_data.schema.select_by_tag(
        Tags.SEQUENCE
    ).select_by_tag(Tags.CATEGORICAL)
    seq_schema = sequence_testing_data.schema

    target = sequence_testing_data.schema.select_by_tag(Tags.ITEM_ID).column_names[0]
    predict_last = mm.SequencePredictLast(schema=seq_schema, target=target)
    loader = Loader(sequence_testing_data, batch_size=8, shuffle=False)

    query_schema = seq_schema
    output_schema = seq_schema.select_by_name(target)

    d_model = 48
    query_encoder = mm.Encoder(
        mm.InputBlockV2(
            query_schema,
            categorical=mm.Embeddings(
                query_schema.select_by_tag(Tags.CATEGORICAL), sequence_combiner=None
            ),
        ),
        mm.MLPBlock([d_model]),
        GPT2Block(d_model=d_model, n_head=2, n_layer=2),
        tf.keras.layers.Lambda(lambda x: tf.reduce_mean(x, axis=1)),
    )

    model = mm.RetrievalModelV2(
        query=query_encoder,
        output=mm.ContrastiveOutput(output_schema, negative_samplers="in-batch"),
    )

    testing_utils.model_test(
        model,
        loader,
        run_eagerly=run_eagerly,
        reload_model=False,
        metrics={},
        fit_kwargs={"pre": predict_last},
    )

    predictions = model.predict(loader)
    assert list(predictions.shape) == [100, 51997]

    query_embeddings = query_encoder.predict(loader)
    assert list(query_embeddings.shape) == [100, d_model]

    item_embeddings = model.candidate_embeddings().compute().to_numpy()

    assert list(item_embeddings.shape) == [51997, d_model]
    predicitons_2 = np.dot(query_embeddings, item_embeddings.T)

    np.testing.assert_allclose(predictions, predicitons_2, atol=1e-7)


def test_transformer_encoder():
    NUM_ROWS = 100
    SEQ_LENGTH = 10
    EMBED_DIM = 128

    inputs = tf.random.uniform((NUM_ROWS, SEQ_LENGTH, EMBED_DIM))
    transformer_encod = mm.TransformerBlock(
        transformer=BertConfig(hidden_size=EMBED_DIM, num_attention_heads=16)
    )
    outputs = transformer_encod(inputs)

    assert list(outputs.shape) == [NUM_ROWS, SEQ_LENGTH, EMBED_DIM]


def test_transformer_encoder_with_pooling():
    NUM_ROWS = 100
    SEQ_LENGTH = 10
    EMBED_DIM = 128

    inputs = tf.random.uniform((NUM_ROWS, SEQ_LENGTH, EMBED_DIM))
    transformer_encod = mm.TransformerBlock(
        transformer=BertConfig(hidden_size=EMBED_DIM, num_attention_heads=16),
        transformer_post="pooler_output",
    )
    outputs = transformer_encod(inputs)

    assert list(outputs.shape) == [NUM_ROWS, EMBED_DIM]


@pytest.mark.parametrize("max_seq_length", [None, 5, 20])
def test_transformer_encoder_with_list_to_dense(max_seq_length):
    NUM_ROWS = 100
    SEQ_LENGTH = 10
    EMBED_DIM = 128
    inputs = tf.RaggedTensor.from_tensor(tf.random.uniform((NUM_ROWS, SEQ_LENGTH, EMBED_DIM)))

    transformer_encod = mm.TransformerBlock(
        transformer=BertConfig(hidden_size=EMBED_DIM, num_attention_heads=16),
        pre=mm.ListToDense(max_seq_length=max_seq_length),
    )
    outputs = transformer_encod(inputs)

    if max_seq_length is not None:
        assert list(outputs.shape) == [NUM_ROWS, max_seq_length, EMBED_DIM]
    else:
        assert list(outputs.shape) == [NUM_ROWS, SEQ_LENGTH, EMBED_DIM]


def test_transformer_encoder_with_post():
    NUM_ROWS = 100
    SEQ_LENGTH = 10
    EMBED_DIM = 128
    inputs = tf.RaggedTensor.from_tensor(tf.random.uniform((NUM_ROWS, SEQ_LENGTH, EMBED_DIM)))

    transformer_encod = mm.TransformerBlock(
        transformer=BertConfig(hidden_size=EMBED_DIM, num_attention_heads=16),
        pre=mm.ListToDense(max_seq_length=5),
        post="sequence_mean",
    )
    outputs = transformer_encod(inputs)
    testing_utils.assert_serialization(transformer_encod)
    assert list(outputs.shape) == [NUM_ROWS, EMBED_DIM]


@pytest.mark.parametrize("encoder", [XLNetBlock, BertBlock, AlbertBlock, RobertaBlock, GPT2Block])
def test_hf_tranformers_blocks(encoder):
    NUM_ROWS = 100
    SEQ_LENGTH = 10
    EMBED_DIM = 128
    inputs = tf.random.uniform((NUM_ROWS, SEQ_LENGTH, EMBED_DIM))
    transformer_encod = encoder(
        d_model=EMBED_DIM,
        n_head=8,
        n_layer=2,
    )
    outputs = transformer_encod(inputs)
    assert list(outputs.shape) == [NUM_ROWS, SEQ_LENGTH, EMBED_DIM]


@pytest.mark.parametrize("run_eagerly", [True, False])
def test_transformer_as_classification_model(sequence_testing_data: Dataset, run_eagerly):
    EMBED_DIM = 48
    loader, schema = classification_loader(sequence_testing_data)

    model = mm.Model(
        mm.InputBlockV2(
            schema,
            categorical=mm.Embeddings(schema, sequence_combiner=None),
        ),
        BertBlock(
            d_model=EMBED_DIM,
            n_head=8,
            n_layer=2,
            transformer_post="pooler_output",
        ),
        mm.CategoricalOutput(
            to_call=schema["user_country"],
        ),
    )

    batch = loader.peek()[0]

    outputs = model(batch)
    assert list(outputs.shape) == [50, 63]
    testing_utils.model_test(model, loader, run_eagerly=run_eagerly)


def test_tranformer_with_prepare_module(sequence_testing_data):
    NUM_ROWS = 100
    SEQ_LENGTH = 10
    EMBED_DIM = 128
    inputs = tf.random.uniform((NUM_ROWS, SEQ_LENGTH, EMBED_DIM))

    class DummyPrepare(tf.keras.layers.Layer):
        def __init__(self, transformer, **kwargs):
            self.transformer = transformer
            super().__init__(**kwargs)

        def call(self, inputs, features=None):
            bs = tf.shape(inputs)[0]
            seq_len = self.transformer.config.max_position_embeddings
            attention_mask = tf.ones((bs, seq_len))
            inputs = {"inputs_embeds": inputs, "attention_mask": attention_mask}
            return inputs

    transformer_encod = BertBlock(
        d_model=EMBED_DIM,
        n_head=8,
        n_layer=2,
        max_position_embeddings=SEQ_LENGTH,
        transformer_pre=DummyPrepare,
    )

    outputs = transformer_encod(inputs)
    assert list(outputs.shape) == [NUM_ROWS, SEQ_LENGTH, EMBED_DIM]


def classification_loader(sequence_testing_data: Dataset):
    schema = sequence_testing_data.schema.select_by_name(
        ["item_id_seq", "categories", "user_country"]
    )
    sequence_testing_data.schema = schema
    dataloader = mm.Loader(
        sequence_testing_data,
        batch_size=50,
    ).map(mm.ToTarget(schema, "user_country", one_hot=True))
    return dataloader, schema


@pytest.mark.parametrize("run_eagerly", [True, False])
def test_transformer_with_causal_language_modeling(sequence_testing_data: Dataset, run_eagerly):

    seq_schema = sequence_testing_data.schema.select_by_tag(Tags.SEQUENCE).select_by_tag(
        Tags.CATEGORICAL
    )
    target = sequence_testing_data.schema.select_by_tag(Tags.ITEM_ID).column_names[0]
    predict_next = mm.SequencePredictNext(schema=seq_schema, target=target)

    loader = Loader(sequence_testing_data, batch_size=8, shuffle=False)

    model = mm.Model(
        mm.InputBlockV2(
            seq_schema,
            categorical=mm.Embeddings(
                seq_schema.select_by_tag(Tags.CATEGORICAL), sequence_combiner=None
            ),
        ),
        GPT2Block(d_model=48, n_head=8, n_layer=2),
        mm.CategoricalOutput(
            seq_schema.select_by_name(target), default_loss="categorical_crossentropy"
        ),
    )

    batch = next(iter(loader))[0]
    outputs = model(batch)
    assert list(outputs.shape) == [8, 4, 51997]
    testing_utils.model_test(
        model, loader, run_eagerly=run_eagerly, reload_model=True, fit_kwargs={"pre": predict_next}
    )

    metrics = model.evaluate(loader, batch_size=8, steps=1, return_dict=True, pre=predict_next)
    assert len(metrics) > 0

    predictions = model.predict(loader, batch_size=8, steps=1)
    assert predictions.shape == (8, 4, 51997)


@pytest.mark.parametrize("run_eagerly", [True, False])
def test_transformer_with_masked_language_modeling(sequence_testing_data: Dataset, run_eagerly):

    seq_schema = sequence_testing_data.schema.select_by_tag(Tags.SEQUENCE).select_by_tag(
        Tags.CATEGORICAL
    )
    target = sequence_testing_data.schema.select_by_tag(Tags.ITEM_ID).column_names[0]

    loader = Loader(sequence_testing_data, batch_size=8, shuffle=False)
    model = mm.Model(
        mm.InputBlockV2(
            seq_schema,
            categorical=mm.Embeddings(
                seq_schema.select_by_tag(Tags.CATEGORICAL), sequence_combiner=None
            ),
        ),
        BertBlock(
            d_model=48,
            n_head=8,
            n_layer=2,
            pre=mm.SequentialBlock([mm.SequenceMaskLastInference(), mm.ReplaceMaskedEmbeddings()]),
            post="inference_hidden_state",
        ),
        mm.CategoricalOutput(
            seq_schema.select_by_name(target),
            default_loss="categorical_crossentropy",
        ),
    )
    seq_mask_random = mm.SequenceMaskRandom(schema=seq_schema, target=target, masking_prob=0.3)

    inputs, targets = loader.peek()

    outputs = model(inputs, targets=targets, training=True)
    assert list(outputs.shape) == [8, 4, 51997]
    testing_utils.model_test(
        model,
        loader,
        run_eagerly=run_eagerly,
        reload_model=True,
        fit_kwargs={"pre": seq_mask_random},
    )

    seq_mask_last = mm.SequenceMaskLast(schema=seq_schema, target=target)
    metrics = model.evaluate(loader, batch_size=8, steps=1, return_dict=True, pre=seq_mask_last)
    assert len(metrics) > 0

    # Get predictions for next-item position
    predictions = model.predict(loader, batch_size=8, steps=1)
    assert predictions.shape == (8, 51997)


@pytest.mark.parametrize("run_eagerly", [True, False])
def test_transformer_with_masked_language_modeling_check_eval_masked(
    sequence_testing_data: Dataset, run_eagerly
):

    seq_schema = sequence_testing_data.schema.select_by_tag(Tags.SEQUENCE).select_by_tag(
        Tags.CATEGORICAL
    )
    target = sequence_testing_data.schema.select_by_tag(Tags.ITEM_ID).column_names[0]

    loader = Loader(sequence_testing_data, batch_size=8, shuffle=False)
    model = mm.Model(
        mm.InputBlockV2(
            seq_schema,
            categorical=mm.Embeddings(
                seq_schema.select_by_tag(Tags.CATEGORICAL), sequence_combiner=None
            ),
        ),
        # BertBlock(d_model=48, n_head=8, n_layer=2, pre=mm.ReplaceMaskedEmbeddings()),
        GPT2Block(d_model=48, n_head=8, n_layer=2, pre=mm.ReplaceMaskedEmbeddings()),
        mm.CategoricalOutput(
            seq_schema.select_by_name(target),
            default_loss="categorical_crossentropy",
        ),
    )
    seq_mask_random = mm.SequenceMaskRandom(schema=seq_schema, target=target, masking_prob=0.3)

    inputs = itertools.islice(iter(loader), 1)
    outputs = model.predict(inputs, pre=seq_mask_random)
    assert list(outputs.shape) == [8, 4, 51997]

    testing_utils.model_test(
        model,
        loader,
        run_eagerly=run_eagerly,
        reload_model=True,
        fit_kwargs={"pre": seq_mask_random},
        metrics=[mm.RecallAt(5000), mm.NDCGAt(5000, seed=4)],
    )

    # This transform only extracts targets, but without applying mask
    seq_target_as_input_no_mask = mm.SequenceTargetAsInput(schema=seq_schema, target=target)

    with Loader(sequence_testing_data, batch_size=8, shuffle=False) as loader:
        metrics_all_positions1 = model.evaluate(
            loader, batch_size=8, steps=1, return_dict=True, pre=seq_target_as_input_no_mask
        )

    with Loader(sequence_testing_data, batch_size=8, shuffle=False) as loader:
        metrics_all_positions2 = model.evaluate(
            loader, batch_size=8, steps=1, return_dict=True, pre=seq_target_as_input_no_mask
        )

    def _metrics_almost_equal(metrics1, metrics2):
        return np.all(
            [
                np.isclose(metrics1[k], metrics2[k], atol=1e-05)
                for k in metrics1
                if k not in "regularization_loss"
            ]
        )

    # Ensures metrics without masked positions are equal
    assert _metrics_almost_equal(metrics_all_positions1, metrics_all_positions2)

    seq_mask_last = mm.SequenceMaskLast(schema=seq_schema, target=target)
    metrics_last_positions = model.evaluate(
        loader, batch_size=8, steps=1, return_dict=True, pre=seq_mask_last
    )
    # Ensures metrics masking only last positions are different then the ones
    # considering all positions
    assert not _metrics_almost_equal(metrics_all_positions1, metrics_last_positions)



import os

import pytest
import tensorflow as tf
from tensorflow.keras.utils import set_random_seed

import merlin.models.tf as mm
from merlin.datasets.advertising.criteo.dataset import default_criteo_transform
from merlin.io.dataset import Dataset
from merlin.models.tf.distributed.backend import hvd, hvd_installed
from merlin.models.utils.example_utils import workflow_fit_transform
from merlin.schema.tags import Tags

# Set seed to make tests deterministic. In real training, it is not
# necessary to set the random seed.
set_random_seed(42)


@pytest.mark.parametrize("custom_distributed_optimizer", [True, False])
def test_horovod_multigpu_dlrm(
    criteo_data,
    tmpdir,
    custom_distributed_optimizer,
    batch_size=11,
    learning_rate=0.03,
):
    """
    This test should work on CPU, single GPU, and multiple GPUs.
    However, for distributed training on multiple GPUs, it needs to be
    executed with `horovodrun`:

    $ horovodrun -np 2 hvd_wrapper.sh python -m pytest tests/unit/tf/horovod/test_horovod.py

    Even with multiple GPUs, if you simply run pytest without horovodrun, it
    will use only one GPU.
    """
    criteo_data.to_ddf().to_parquet(os.path.join(tmpdir, "train"))
    criteo_data.to_ddf().to_parquet(os.path.join(tmpdir, "valid"))
    workflow = default_criteo_transform()
    workflow_fit_transform(
        workflow,
        os.path.join(tmpdir, "train", "*.parquet"),
        os.path.join(tmpdir, "valid", "*.parquet"),
        os.path.join(tmpdir, "processed"),
    )

    # As a workaround for nvtabular producing different numbers of batches in
    # workers, we repartition the dataset in multiples of hvd.size().
    # https://github.com/NVIDIA-Merlin/models/issues/765
    train = Dataset(os.path.join(tmpdir, "processed", "train", "*.parquet"))
    ddf = train.to_ddf().repartition(npartitions=2)
    train = Dataset(ddf, schema=train.schema)

    train_loader = mm.Loader(
        train,
        schema=train.schema,
        batch_size=batch_size,
        shuffle=True,
        drop_last=True,
    )

    target_column = train.schema.select_by_tag(Tags.TARGET).column_names[0]

    model = mm.DLRMModel(
        train.schema,
        embedding_dim=16,
        bottom_block=mm.MLPBlock([32, 16]),
        top_block=mm.MLPBlock([32, 16]),
        prediction_tasks=mm.BinaryClassificationTask(target_column),
    )

    if custom_distributed_optimizer:
        # Test for a case when the user uses a custom DistributedOptimizer.
        # With a custom hvd.DistributedOptimzer, users have to adjust the learning rate.
        opt = tf.keras.optimizers.Adagrad(learning_rate=learning_rate * hvd.size())
        opt = hvd.DistributedOptimizer(opt, compression=hvd.Compression.fp16)
    else:
        opt = tf.keras.optimizers.Adagrad(learning_rate=learning_rate)

    model.compile(optimizer=opt, run_eagerly=False, metrics=[tf.keras.metrics.AUC()])

    # model.fit() will hang or terminate with error if all workers don't have
    # the same number of batches.
    losses = model.fit(
        train_loader,
        batch_size=batch_size,
    )

    model.save(tmpdir)

    if hvd_installed:
        assert model.optimizer.learning_rate == learning_rate * hvd.size()
    else:
        assert model.optimizer.learning_rate == learning_rate

    assert all(measure >= 0 for metric in losses.history for measure in losses.history[metric])

    # Check the steps in each worker to check that the dataset is distributed
    # across workers. If this works correctly in a multi-gpu setting, the steps
    # should decrease with more workers, e.g., steps = 9 in each worker with
    # 1 GPU, steps = 4 in each worker with 2 GPUS, steps = 3 in each worker
    # with 3 GPUS, and so on.
    if hvd_installed:
        assert losses.params["steps"] == 9 // hvd.size()
    else:
        assert losses.params["steps"] == 9

    saved_model = "saved_model.pb"
    if hvd_installed and hvd.rank() == 0:
        assert saved_model in os.listdir(tmpdir)
    if hvd_installed and hvd.rank() != 0:
        assert saved_model not in os.listdir(tmpdir)


def test_horovod_multigpu_two_tower(
    music_streaming_data, tmpdir, batch_size=11, learning_rate=0.03
):
    """
    This test should work on CPU, single GPU, and multiple GPUs.
    However, for distributed training on multiple GPUs, it needs to be
    executed with `horovodrun`:

    $ horovodrun -np 2 hvd_wrapper.sh python -m pytest tests/unit/tf/horovod/test_horovod.py

    Even with multiple GPUs, if you simply run pytest without horovodrun, it
    will use only one GPU.
    """
    # As a workaround for nvtabular producing different numbers of batches in
    # workers, we repartition the dataset in multiples of hvd.size().
    # https://github.com/NVIDIA-Merlin/models/issues/765
    music_streaming_data.schema = music_streaming_data.schema.select_by_name(
        ["item_id", "user_genres"]
    )
    ddf = music_streaming_data.to_ddf().repartition(npartitions=2)
    train = Dataset(ddf, schema=music_streaming_data.schema)

    train_loader = mm.Loader(
        train,
        schema=train.schema,
        batch_size=batch_size,
        shuffle=True,
        drop_last=True,
    )

    model = mm.TwoTowerModel(music_streaming_data.schema, query_tower=mm.MLPBlock([2]))
    model.compile(optimizer="adam", run_eagerly=False)

    # model.fit() will hang or terminate with error if all workers don't have
    # the same number of batches.
    losses = model.fit(train_loader, batch_size=batch_size, epochs=2)

    model.save(tmpdir)

    assert all(measure >= 0 for metric in losses.history for measure in losses.history[metric])

    if hvd_installed:
        assert losses.params["steps"] == 9 // hvd.size()  # 9 steps per epoch; 2 epochs
    else:
        assert losses.params["steps"] == 9

    saved_model = "saved_model.pb"
    if hvd_installed and hvd.rank() != 0:
        assert saved_model not in os.listdir(tmpdir)
    else:
        assert saved_model in os.listdir(tmpdir)



import os
import urllib
from pathlib import Path
from typing import Optional, Tuple, Union

import numpy as np
from tqdm import tqdm

import merlin.io
from merlin.core.dispatch import get_lib
from merlin.dag import ColumnSelector
from merlin.datasets import BASE_PATH
from merlin.models.utils.example_utils import workflow_fit_transform
from merlin.models.utils.nvt_utils import require_nvt
from merlin.schema import Tags

try:
    import nvtabular as nvt

    Workflow = nvt.Workflow
except ImportError:
    Workflow = None


_FILES = ["ground_truth.csv", "test_set.csv", "train_set.csv"]
_DATA_URL = "https://raw.githubusercontent.com/bookingcom/ml-dataset-mdt/main/"


def get_booking(
    path: Optional[Union[str, Path]] = None,
    overwrite: bool = False,
    transformed_name: str = "transformed",
    nvt_workflow: Optional[Workflow] = None,
    **kwargs,
) -> Tuple[merlin.io.Dataset, merlin.io.Dataset]:
    """Dataset for the WSDM '21 challenge organized by booking.com.

    The goal of this challenge is to use a dataset based on
    millions of real anonymized accommodation reservations
    to come up with a strategy for making the best recommendation
    for their next destination in real-time.

    Parameters
    ----------
    path (Union[str, Path]): Path to save the dataset.
    overwrite (bool, optional): Whether or not to overwrite the dataset,
        if already downloaded. Defaults to False.
    transformed_name (str, optional): Name of folder to put the transformed dataset in.
        Defaults to "transformed".
    nvt_workflow (Optional[Workflow], optional): NVTabular workflow, pass this in
        if you want to customize the default workflow. Defaults to `default_booking_transformation`.

    Returns
    -------
        train: merlin.io.Dataset
            Training dataset.
        valid: merlin.io.Dataset
            Test dataset.
    """
    require_nvt()

    if path is None:
        p = Path(BASE_PATH) / "booking"
    else:
        p = Path(path)

    raw_path = p
    if not (
        _check_path(raw_path / "train", check_schema=False)
        and _check_path(raw_path / "test", check_schema=False)
    ):
        download_booking(p)

    nvt_path = raw_path / transformed_name
    train_path, valid_path = nvt_path / "train", nvt_path / "valid"
    nvt_path_exists = _check_path(train_path) and _check_path(valid_path)
    if not nvt_path_exists or overwrite:
        transform_booking(raw_path, nvt_path, nvt_workflow=nvt_workflow, **kwargs)

    _check_path(train_path, strict=True)
    _check_path(valid_path, strict=True)
    train = merlin.io.Dataset(str(train_path), engine="parquet")
    valid = merlin.io.Dataset(str(valid_path), engine="parquet")

    return train, valid


def download_booking(path: Path):
    """Automatically download the booking dataset.

    Parameters
    ----------
    path (Path): output-path.
    """
    path = Path(path)
    path.mkdir(parents=True, exist_ok=True)

    for file in _FILES:
        local_filename = str(path / file)
        url = os.path.join(_DATA_URL, file)
        desc = f"downloading {os.path.basename(local_filename)}"
        with tqdm(unit="B", unit_scale=True, desc=desc) as progress:

            def report(chunk, chunksize, total):
                if not progress.total:
                    progress.reset(total=total)
                progress.update(chunksize)

            urllib.request.urlretrieve(url, local_filename, reporthook=report)

    preprocess_booking(path)


def preprocess_booking(
    path: Path,
):
    path = Path(path)
    train = get_lib().read_csv(path / "train_set.csv")
    test = get_lib().read_csv(path / "test_set.csv")

    train["checkin"] = get_lib().to_datetime(train["checkin"], format="%Y-%m-%d")
    train["checkout"] = get_lib().to_datetime(train["checkout"], format="%Y-%m-%d")
    train["timestamp"] = train["checkout"].astype("int64")
    test["checkin"] = get_lib().to_datetime(test["checkin"], format="%Y-%m-%d")
    test["checkout"] = get_lib().to_datetime(test["checkout"], format="%Y-%m-%d")
    test["timestamp"] = test["checkout"].astype("int64")

    (path / "train").mkdir(exist_ok=True)
    (path / "test").mkdir(exist_ok=True)

    train.to_parquet(path / "train/data.parquet")
    test.to_parquet(path / "test/data.parquet")


def _get_cycled_feature_value_sin(col, max_value=7):
    value_scaled = (col + 0.000001) / max_value
    value_sin = np.sin(2 * np.pi * value_scaled)
    return value_sin


def _get_cycled_feature_value_cos(col, max_value=7):
    value_scaled = (col + 0.000001) / max_value
    value_cos = np.cos(2 * np.pi * value_scaled)
    return value_cos


def transform_booking(
    data: Union[str, Path, Tuple[merlin.io.Dataset, merlin.io.Dataset]],
    output_path: Union[str, Path],
    nvt_workflow=None,
    **kwargs,
):
    """Transform the booking dataset.

    Parameters
    ----------
    data (Union[str, Path, Tuple[merlin.io.Dataset, merlin.io.Dataset]]):
        Raw training and test data.
    output_path (Union[str, Path]): Path to save the transformed dataset.
    nvt_workflow (Optional[Workflow], optional): NVTabular workflow, pass this in
        if you want to customize the default workflow. Defaults to `default_booking_transformation`.
    """
    nvt_workflow = nvt_workflow or default_booking_transformation(**locals())

    if isinstance(data, (str, Path)):
        _train = merlin.io.Dataset(str(Path(data) / "train"), engine="parquet")
        _valid = merlin.io.Dataset(str(Path(data) / "test"), engine="parquet")
    elif (
        isinstance(data, tuple)
        and len(data) == 2
        and all(isinstance(x, merlin.io.Dataset) for x in data)
    ):
        _train, _valid = data
    else:
        raise ValueError("data must be a path or a tuple of train and valid datasets")

    workflow_fit_transform(nvt_workflow, _train, _valid, str(output_path), **kwargs)


def default_booking_transformation(**kwargs):
    """Default transformation for the booking dataset.

    Returns:
        Workflow: NVTabular workflow
    """
    cat = lambda: nvt.ops.Categorify(start_index=1)  # noqa: E731

    df_season = get_lib().DataFrame(
        {"month": range(1, 13), "season": ([0] * 3) + ([1] * 3) + ([2] * 3) + ([3] * 3)}
    )

    month = (
        ["checkin"]
        >> nvt.ops.LambdaOp(lambda col: col.dt.month.astype("int64"))
        >> nvt.ops.Rename(name="month")
    )

    weekday_checkin = (
        ["checkin"]
        >> nvt.ops.LambdaOp(lambda col: col.dt.weekday)
        >> nvt.ops.Rename(name="weekday_checkin")
    )

    weekday_checkout = (
        ["checkout"]
        >> nvt.ops.LambdaOp(lambda col: col.dt.weekday)
        >> nvt.ops.Rename(name="weekday_checkout")
    )

    is_weekend = (
        weekday_checkin
        >> nvt.ops.LambdaOp(lambda col: col.isin([5, 6]).astype(int))
        >> nvt.ops.Rename(name="is_weekend")
    )

    length_feature = (
        ["checkout"]
        >> nvt.ops.LambdaOp(lambda col, df: (col - df["checkin"]).dt.days, dependency=["checkin"])
        >> nvt.ops.Rename(name="length")
    )

    season = month >> nvt.ops.JoinExternal(
        df_season,
        on="month",
        how="left",
    )

    weekday_sin = (
        weekday_checkout
        >> (lambda col: _get_cycled_feature_value_sin(col + 1, 7))
        >> nvt.ops.Rename(name="dayofweek_sin")
        >> nvt.ops.AddTags([Tags.CONTINUOUS])
    )
    weekday_cos = (
        weekday_checkout
        >> (lambda col: _get_cycled_feature_value_cos(col + 1, 7))
        >> nvt.ops.Rename(name="dayofweek_cos")
        >> nvt.ops.AddTags([Tags.CONTINUOUS])
    )

    context_cat_features = ["device_class", "affiliate_id"] + season >> cat()
    seq_cat_features = (
        ["booker_country", "hotel_country"]
        + month
        + weekday_checkin
        + weekday_checkout
        + weekday_sin
        + weekday_cos
        + is_weekend
        + length_feature
        >> cat()
        >> nvt.ops.AddTags(tags=[Tags.SEQUENCE])
    )

    cityid = ["city_id"] >> cat() >> nvt.ops.AddTags(tags=[Tags.ITEM_ID])
    session_id = (
        ["utrip_id"]
        >> cat()
        >> nvt.ops.AddTags(tags=[Tags.SESSION_ID])
        >> nvt.ops.Rename(name="session_id")
    )
    user_id = ["user_id"] >> cat() >> nvt.ops.AddTags(tags=[Tags.USER_ID])

    features = (
        ColumnSelector(["timestamp", "utrip_id"])
        + seq_cat_features
        + cityid
        + context_cat_features
        + session_id
        + user_id
    )

    grouped = (
        features
        >> nvt.ops.Groupby(
            groupby_cols=["utrip_id"],
            sort_cols=["timestamp"],
            aggs={
                "user_id": "first",
                "session_id": "first",
                "device_class": "first",
                "booker_country": "list",
                "hotel_country": "list",
                "month": "list",
                "is_weekend": "list",
                "weekday_checkin": "list",
                "weekday_checkout": "list",
                "length_feature": "first",
                "city_id": "list",
            },
        )
        >> nvt.ops.Rename(_remove_list_and_first_from_name)
        >> nvt.ops.ValueCount()
    )

    return grouped


def _remove_list_and_first_from_name(name):
    if name.endswith("_list"):
        return name[:-5]

    if name.endswith("_first"):
        return name[:-6]

    return name


def _check_path(path, strict=False, check_schema=True):
    if not isinstance(path, (str, Path)):
        if strict:
            raise ValueError("path must be a string or a Path object")

        return False

    if not Path(path).exists():
        if strict:
            raise ValueError(f"path {path} does not exist")

        return False

    if not len(list(Path(path).glob("*.parquet"))):
        if strict:
            raise ValueError(f"path {path} does not contain any parquet files")

        return False

    if check_schema:
        if not len(list(Path(path).glob("schema.*"))):
            if strict:
                raise ValueError(f"path {path} does not contain a schema")

            return False

    return True



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fiddle as fdl

from merlin.io import Dataset
from tests.common.tf.retrieval.retrieval_utils import (
    RetrievalTrainEvalRunner,
    RetrievalTrainEvalRunnerV2,
    WandbLogger,
    filter_schema,
    get_callbacks,
    get_dual_encoder_model,
    get_dual_encoder_model_v2,
    get_item_frequencies,
    get_loss,
    get_metrics,
    get_optimizer,
    get_samplers,
    get_samplers_v2,
    get_youtube_dnn_model,
)


def make_model(schema, train_ds, model_type):
    if model_type == "youtubednn":
        model = fdl.Config(get_youtube_dnn_model, schema)
    else:
        samplers = fdl.Config(get_samplers, schema)
        items_frequencies = fdl.Config(get_item_frequencies, schema, train_ds)
        model = fdl.Config(get_dual_encoder_model, schema, samplers, items_frequencies, model_type)
    return model


def make_model_v2(schema, train_ds, model_type):
    samplers = fdl.Config(get_samplers_v2, schema)
    items_frequencies = fdl.Config(get_item_frequencies, schema, train_ds)
    model = fdl.Config(get_dual_encoder_model_v2, schema, samplers, items_frequencies, model_type)
    return model


def config_retrieval_train_eval_runner(
    train_ds: Dataset,
    eval_ds: Dataset,
    model_type: str,
    log_to_wandb: bool,
    wandb_project: str,
    retrieval_api_version: int,
):
    wandb_logger_cfg = fdl.Config(WandbLogger, enabled=log_to_wandb, wandb_project=wandb_project)

    schema_cfg = fdl.Config(filter_schema, schema=train_ds.schema)
    optimizer = fdl.Config(get_optimizer)
    metrics = fdl.Config(get_metrics)
    loss = fdl.Config(get_loss)
    callbacks = fdl.Config(get_callbacks, wandb_logger=wandb_logger_cfg)
    if retrieval_api_version == 1:
        model_cfg = make_model(schema_cfg, train_ds, model_type=model_type)
        runner_cfg = fdl.Config(
            RetrievalTrainEvalRunner,
            wandb_logger=wandb_logger_cfg,
            model_type=model_type,
            schema=schema_cfg,
            train_ds=train_ds,
            eval_ds=eval_ds,
            model=model_cfg,
            optimizer=optimizer,
            metrics=metrics,
            loss=loss,
            callbacks=callbacks,
        )

    elif retrieval_api_version == 2:
        model_cfg = make_model_v2(schema_cfg, train_ds, model_type=model_type)
        runner_cfg = fdl.Config(
            RetrievalTrainEvalRunnerV2,
            wandb_logger=wandb_logger_cfg,
            model_type=model_type,
            schema=schema_cfg,
            train_ds=train_ds,
            eval_ds=eval_ds,
            model=model_cfg,
            optimizer=optimizer,
            metrics=metrics,
            loss=loss,
            callbacks=callbacks,
        )

    return runner_cfg



import os

from testbook import testbook

from tests.conftest import REPO_ROOT


@testbook(REPO_ROOT / "examples/04-Exporting-ranking-models.ipynb", execute=False)
def test_example_04_exporting_ranking_models(tb, tmpdir):
    tb.inject(
        f"""
        import os
        os.environ["DATA_FOLDER"] = "{tmpdir}"
        os.environ["NUM_ROWS"] = "999"
        """
    )
    tb.execute()
    assert os.path.isdir(os.path.join(tmpdir, "dlrm"))
    assert os.path.isdir(os.path.join(tmpdir, "workflow"))



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import merlin.models.tf as mm
from merlin.schema import Tags


def test_concat_sequence(sequence_testing_data):
    sequence_testing_data.schema = sequence_testing_data.schema.select_by_tag(Tags.SEQUENCE)
    seq_schema = sequence_testing_data.schema

    seq_inputs = mm.InputBlockV2(
        seq_schema,
        aggregation="concat",
        categorical=mm.Embeddings(
            seq_schema.select_by_tag(Tags.CATEGORICAL), sequence_combiner=None
        ),
    )

    inputs = mm.sample_batch(sequence_testing_data, 8, include_targets=False, process_lists=True)

    outputs = seq_inputs(inputs)

    assert outputs.shape.rank == 3



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Dict, NamedTuple, Optional, Union

import tensorflow as tf

from merlin.core.dispatch import DataFrameType, get_lib

TensorLike = Union[tf.Tensor, tf.SparseTensor, tf.RaggedTensor]


class PredictionContext(NamedTuple):
    features: Dict[str, TensorLike]
    targets: Optional[Union[tf.Tensor, Dict[str, tf.Tensor]]] = None
    training: bool = False
    testing: bool = False

    def with_updates(
        self, targets=None, features=None, training=None, testing=None
    ) -> "PredictionContext":
        return PredictionContext(
            features if features is not None else self.features,
            targets if targets is not None else self.targets,
            training or self.training,
            testing or self.testing,
        )

    def to_call_dict(self):
        outputs = {
            "features": self.features,
            "training": self.training,
            "testing": self.testing,
        }

        if self.training or self.testing:
            outputs["targets"] = self.targets

        return outputs


class Prediction(NamedTuple):
    """Prediction object returned by `ModelOutput` classes"""

    outputs: Dict[str, TensorLike]
    targets: Optional[Union[tf.Tensor, Dict[str, tf.Tensor]]] = None
    sample_weight: Optional[tf.Tensor] = None
    features: Optional[Dict[str, TensorLike]] = None
    negative_candidate_ids: Optional[tf.Tensor] = None

    @property
    def predictions(self):
        return self.outputs

    def copy_with_updates(
        self,
        outputs=None,
        targets=None,
        sample_weight=None,
        features=None,
        negative_candidate_ids=None,
    ) -> "Prediction":
        return Prediction(
            outputs if outputs is not None else self.outputs,
            targets if targets is not None else self.targets,
            sample_weight if sample_weight is not None else self.sample_weight,
            features if features is not None else self.features,
            negative_candidate_ids
            if negative_candidate_ids is not None
            else self.negative_candidate_ids,
        )


class TopKPrediction(NamedTuple):
    """Prediction object returned by `TopKOutput` classes"""

    scores: tf.Tensor
    identifiers: tf.Tensor

    def to_df(self) -> DataFrameType:
        """Convert Top-k scores and identifiers to a data-frame."""
        score_names = [f"score_{i}" for i in range(self.scores.shape[1])]
        id_names = [f"id_{i}" for i in range(self.identifiers.shape[1])]

        rows = []
        for batch_i in range(self.scores.shape[0]):
            row = {}
            for k in range(self.scores.shape[1]):
                row[score_names[k]] = encode_output(self.scores[batch_i, k])
                row[id_names[k]] = encode_output(self.identifiers[batch_i, k])
            rows.append(row)

        return get_lib().DataFrame(rows)

    @staticmethod
    def output_names(k: int):
        """Set column names of scores and identifiers when
        `to_df` is called
        """
        score_names = [f"score_{i}" for i in range(k)]
        id_names = [f"id_{i}" for i in range(k)]

        return score_names + id_names


def encode_output(output: tf.Tensor):
    if len(output.shape) == 2 and output.shape[1] == 1:
        output = tf.squeeze(output)

    return output.numpy()



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from merlin.datasets.entertainment.movielens.dataset import (
    default_ml1m_transformation,
    default_ml25m_transformation,
    default_ml100k_transformation,
    get_movielens,
    transform_movielens,
)

__all__ = [
    "get_movielens",
    "transform_movielens",
    "default_ml100k_transformation",
    "default_ml1m_transformation",
    "default_ml25m_transformation",
]



from testbook import testbook

from tests.conftest import REPO_ROOT


@testbook(
    REPO_ROOT / "examples/usecases/multi-gpu-data-parallel-training.ipynb",
    execute=False,
    timeout=180,
)
def test_usecase_data_parallel(tb, tmpdir):
    tb.inject(
        f"""
        import os
        os.environ["DATA_FOLDER"] = "{tmpdir}"
        os.environ["NUM_GPUs"]="1"
        """
    )
    tb.execute()



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Dict, List, Optional, Tuple, Union

import tensorflow as tf

from merlin.models.tf.core.aggregation import StackFeatures
from merlin.models.tf.core.base import Block
from merlin.models.tf.core.combinators import (
    ParallelBlock,
    SequentialBlock,
    TabularBlock,
    WithShortcut,
)
from merlin.models.tf.prediction_tasks.base import ParallelPredictionBlock, PredictionTask
from merlin.models.tf.typing import TabularData
from merlin.models.tf.utils.tf_utils import (
    maybe_deserialize_keras_objects,
    maybe_serialize_keras_objects,
)


@tf.keras.utils.register_keras_serializable(package="merlin_models")
class ExpertsGate(Block):
    """MMoE Gate, which uses input features to generate softmax weights
    in a weighted sum of expert outputs.

    Parameters
    ----------
    num_experts : int
        Number of experts, so that there is a weight for each expert
    gate_block : Block, optional
        Allows for having a Block (e.g. MLPBlock([32])) to combine the inputs
        before the final projection layer (created automatically)
        that outputs a softmax distribution over the number of experts.
        This might give more capacity to the gates to decide from the inputs
        how to better combine the experts.
    softmax_temperature : float, optional
        The temperature of the softmax that is used for weighting the experts outputs,
        by default 1.0
    enable_gate_weights_metric : bool, optional
        Enables logging the average gate weights on experts
    name : str, optional
        Name of the block, by default None
    """

    def __init__(
        self,
        num_experts: int,
        gate_block: Optional[Block] = None,
        softmax_temperature: float = 1.0,
        enable_gate_weights_metrics: bool = False,
        name: str = None,
        **kwargs,
    ):
        super().__init__(name=name, **kwargs)
        self.num_experts = num_experts
        self.softmax_temperature = softmax_temperature
        self.enable_gate_weights_metrics = enable_gate_weights_metrics
        self.gate_name = name

        self.gate_block = gate_block
        self.gate_final = tf.keras.layers.Dense(
            num_experts, use_bias=False, name=f"gate_final_{name}"
        )

    def call(self, inputs: TabularData, **kwargs) -> tf.Tensor:
        """MMOE call

        Parameters
        ----------
        inputs : TabularData
            Expects a dict/TabularData with "shortcut" (inputs) and "experts" blocks

        Returns
        -------
        tf.Tensor
            The experts outputs weighted by the gate
        """
        if set(inputs.keys()) != set(["shortcut", "experts"]):
            raise ValueError("MMoE gate expects a dict with 'shortcut' and 'experts' keys.")

        inputs_shortcut, expert_outputs = inputs["shortcut"], inputs["experts"]

        inputs_gate = inputs_shortcut
        if self.gate_block is not None:
            inputs_gate = self.gate_block(inputs_gate)
        gate_weights = tf.expand_dims(
            tf.nn.softmax(self.gate_final(inputs_gate) / self.softmax_temperature),
            axis=-1,
            name="gate_softmax",
        )
        out = tf.reduce_sum(expert_outputs * gate_weights, axis=1, keepdims=False)

        if self.enable_gate_weights_metrics:
            gate_weights_batch_mean = tf.reduce_mean(gate_weights, axis=0, keepdims=False)
            for i in range(gate_weights.shape[1]):
                self.add_metric(gate_weights_batch_mean[i], name=f"{self.gate_name}_weight_{i}")

        return out

    def compute_output_shape(self, input_shape):
        return tf.TensorShape([input_shape["experts"][0], input_shape["experts"][2]])

    @classmethod
    def from_config(cls, config, **kwargs):
        config = maybe_deserialize_keras_objects(config, ["gate_block"])
        return cls(**config)

    def get_config(self):
        config = super().get_config()
        config = maybe_serialize_keras_objects(self, config, ["gate_block"])
        config.update(
            num_experts=self.num_experts,
            softmax_temperature=self.softmax_temperature,
            enable_gate_weights_metrics=self.enable_gate_weights_metrics,
            name=self.gate_name,
        )

        return config


def MMOEBlock(
    outputs: Union[List[str], List[PredictionTask], ParallelPredictionBlock],
    expert_block: Block,
    num_experts: int,
    gate_block: Optional[Block] = None,
    gate_softmax_temperature: float = 1.0,
    **gate_kwargs,
) -> SequentialBlock:
    """Implements the Multi-gate Mixture-of-Experts (MMoE) introduced in [1].

    References
    ----------
    [1] Ma, Jiaqi, et al. "Modeling task relationships in multi-task learning with
    multi-gate mixture-of-experts." Proceedings of the 24th ACM SIGKDD international
    conference on knowledge discovery & data mining. 2018.

    Parameters
    ----------
    outputs : Union[List[str], List[PredictionTask], ParallelPredictionBlock]
        List with the tasks. A gate is created for each task.
    expert_block : Block
        Expert block to be replicated, e.g. MLPBlock([64])
    num_experts : int
        Number of experts to be replicated
    gate_block : Block, optional
        Allows for having a Block (e.g. MLPBlock([32])) to combine the inputs
        before the final projection layer (created automatically)
        that outputs a softmax distribution over the number of experts.
        This might give more capacity to the gates to decide from the inputs
        how to better combine the experts.
    gate_softmax_temperature : float, optional
        The temperature used by the gates, by default 1.0.
        It can be used to smooth the weights distribution over experts outputs.

    Returns
    -------
    SequentialBlock
        Outputs the sequence of blocks that implement MMOE
    """
    if isinstance(outputs, ParallelPredictionBlock):
        output_names = outputs.task_names
    elif all(isinstance(x, PredictionTask) for x in outputs):
        output_names = [o.task_name for o in outputs]  # type: ignore
    else:
        output_names = outputs  # type: ignore

    if not isinstance(expert_block, Block):
        expert_block = Block.from_layer(expert_block)

    experts = expert_block.repeat_in_parallel(
        num_experts, prefix="expert_", aggregation=StackFeatures(axis=1)
    )

    gates = {
        output_name: ExpertsGate(
            num_experts,
            gate_block=gate_block.copy() if gate_block else None,
            softmax_temperature=gate_softmax_temperature,
            name=f"gate_{output_name}",
            **gate_kwargs,
        )
        for output_name in output_names
    }
    gates = ParallelBlock(gates)

    mmoe = WithShortcut(
        experts,
        block_outputs_name="experts",
        automatic_pruning=False,
    )
    mmoe = mmoe.connect(gates, block_name="mmoe")

    return mmoe


@tf.keras.utils.register_keras_serializable(package="merlin_models")
class CGCGateTransformation(TabularBlock):
    """Use the tasks (and shared) gates to aggregate
    the outputs of the experts.

    Parameters
    ----------
    task_names : List[str]
        List with the task names
    num_task_experts : int, optional
        Number of task-specific experts, by default 1
    num_task_experts : int, optional
        Number of task-specific experts, by default 1
    num_shared_experts : int, optional
        Number of shared experts for tasks, by default 1
    add_shared_gate : bool, optional
        Whether to add a shared gate for this CGC block, by default False.
        Useful when multiple CGC blocks are stacked (e.g. in PLEBlock)
        As all CGC blocks except the last one should include the shared gate.
    gate_block : Optional[Block], optional
        Optional block than can make the Gate mode powerful in converting
        the inputs into expert weights for averaging, by default None
    gate_softmax_temperature : float, optional
        Temperature of the softmax used by the Gates for getting
        weights for the average. Temperature can be used to smooth
        the weights distribution, and is by default 1.0.
    enable_gate_weights_metrics : bool, optional
        Enables logging the average gate weights on experts
    gate_dict : Dict[str, ExpertsGate], optional
        Used by the deserializer to rebuild the dict of gates, by default None
    """

    def __init__(
        self,
        task_names: List[str],
        num_task_experts: int = 1,
        num_shared_experts: int = 1,
        add_shared_gate: bool = False,
        gate_block: Optional[Block] = None,
        gate_softmax_temperature: float = 1.0,
        enable_gate_weights_metrics: bool = False,
        gate_dict: Dict[str, ExpertsGate] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        num_total_experts = num_task_experts + num_shared_experts
        self.task_names = list(task_names)  # Creates another list
        if add_shared_gate and "shared" not in self.task_names:
            self.task_names.append("shared")
        self.stack = StackFeatures(axis=1)

        self.gate_dict = gate_dict
        if self.gate_dict is None:
            self.gate_dict: Dict[str, ExpertsGate] = {
                name: ExpertsGate(
                    num_total_experts,
                    gate_block=gate_block.copy() if gate_block else None,
                    softmax_temperature=gate_softmax_temperature,
                    enable_gate_weights_metrics=enable_gate_weights_metrics,
                    name=f"gate_{name}",
                )
                for name in task_names
            }

            if add_shared_gate:
                self.gate_dict["shared"] = ExpertsGate(
                    (len(task_names) * num_task_experts) + num_shared_experts,
                    gate_block=gate_block,
                    softmax_temperature=gate_softmax_temperature,
                    enable_gate_weights_metrics=enable_gate_weights_metrics,
                    name="shared_gate",
                )

    def call(self, expert_outputs: TabularData, **kwargs) -> TabularData:  # type: ignore
        outputs: TabularData = {}

        expert_outputs = dict(expert_outputs)  # Copying the dict
        shortcut = expert_outputs.pop("shortcut")
        outputs["shortcut"] = shortcut

        for name in self.task_names:
            experts = dict(
                experts=self.stack(self.filter_expert_outputs(expert_outputs, name)),
                shortcut=shortcut,
            )
            outputs[name] = self.gate_dict[name](experts)

        return outputs

    def filter_expert_outputs(self, expert_outputs: TabularData, task_name: str) -> TabularData:
        if task_name == "shared":
            return expert_outputs

        filtered_experts: TabularData = {}
        for name, val in expert_outputs.items():
            if name.startswith((task_name, "shared")):
                filtered_experts[name] = val

        return filtered_experts

    def compute_output_shape(self, input_shape):
        tensor_output_shape = list(input_shape.values())[0]

        output_shapes = {name: tensor_output_shape for name in self.task_names}
        return output_shapes

    @classmethod
    def from_config(cls, config, custom_objects=None, **kwargs):
        gate_dict = config.pop("gate_dict")
        if gate_dict is not None:
            gate_dict = {
                name: tf.keras.layers.deserialize(conf, custom_objects=custom_objects)
                for name, conf in gate_dict.items()
            }

        config.update(gate_dict=gate_dict)
        return cls(**config)

    def get_config(self):
        config = super().get_config()
        config = maybe_serialize_keras_objects(self, config, ["gate_dict"])
        config.update(
            task_names=self.task_names,
        )

        return config


@tf.keras.utils.register_keras_serializable(package="merlin_models")
class CGCBlock(ParallelBlock):
    """Implements the Customized Gate Control (CGC) proposed in [1].

    References
    ----------
    [1] Tang, Hongyan, et al. "Progressive layered extraction (ple): A novel multi-task
    learning (mtl) model for personalized recommendations."
    Fourteenth ACM Conference on Recommender Systems. 2020.

    Parameters
    ----------
    outputs : Union[List[str], List[PredictionTask], ParallelPredictionBlock]
       Names of the tasks or PredictionTask/ParallelPredictionBlock objects from
        which we can extract the task names
    expert_block : Union[Block, tf.keras.layers.Layer]
        Block that will be used for the experts
    num_task_experts : int, optional
        Number of task-specific experts, by default 1
    num_shared_experts : int, optional
        Number of shared experts for tasks, by default 1
    add_shared_gate : bool, optional
        Whether to add a shared gate for this CGC block, by default False.
        Useful when multiple CGC blocks are stacked (e.g. in PLEBlock)
        As all CGC blocks except the last one should include the shared gate.
    gate_block : Optional[Block], optional
        Optional block than can make the Gate mode powerful in converting
        the inputs into expert weights for averaging, by default None
    gate_softmax_temperature : float, optional
        Temperature of the softmax used by the Gates for getting
        weights for the average. Temperature can be used to smooth
        the weights distribution, and is by default 1.0.
    enable_gate_weights_metrics : bool, optional
        Enables logging the average gate weights on experts
    name : Optional[str], optional
        Name of the CGC block, by default None
    """

    def __init__(
        self,
        outputs: Union[List[str], List[PredictionTask], ParallelPredictionBlock],
        expert_block: Union[Block, tf.keras.layers.Layer],
        num_task_experts: int = 1,
        num_shared_experts: int = 1,
        add_shared_gate: bool = False,
        gate_block: Optional[Block] = None,
        gate_softmax_temperature: float = 1.0,
        enable_gate_weights_metrics: bool = False,
        name: Optional[str] = None,
        **kwargs,
    ):
        if not isinstance(expert_block, Block):
            expert_block = Block.from_layer(expert_block)

        if isinstance(outputs, ParallelPredictionBlock):
            output_names = outputs.task_names
        elif all(isinstance(x, PredictionTask) for x in outputs):
            output_names = list([o.task_name for o in outputs])  # type: ignore
        else:
            output_names = outputs  # type: ignore
        output_names = list(output_names)
        task_experts = dict(
            [
                create_expert(expert_block, f"{task}/expert_{i}")
                for task in output_names
                for i in range(num_task_experts)
            ]
        )

        shared_experts = dict(
            [create_expert(expert_block, f"shared/expert_{i}") for i in range(num_shared_experts)]
        )

        experts = {**task_experts, **shared_experts}

        post = kwargs.pop("post", None)
        if post is None:
            post = CGCGateTransformation(
                output_names,
                num_task_experts,
                num_shared_experts,
                add_shared_gate=add_shared_gate,
                gate_block=gate_block,
                gate_softmax_temperature=gate_softmax_temperature,
                enable_gate_weights_metrics=enable_gate_weights_metrics,
            )
        super().__init__(
            experts,
            post=post,
            aggregation=None,
            name=name,
            strict=False,
            **kwargs,
        )

    def call(self, inputs, **kwargs):
        if isinstance(inputs, dict):
            outputs = dict(shortcut=inputs["shortcut"])
            for name, layer in self.parallel_dict.items():
                input_name = "/".join(name.split("/")[:-1])
                outputs.update(layer(inputs[input_name]))

            return outputs
        else:
            outputs = super().call(inputs, **kwargs)
            outputs["shortcut"] = inputs  # type: ignore

        return outputs

    def compute_call_output_shape(self, input_shape):
        if isinstance(input_shape, dict):
            output_shapes = {}

            for name, layer in self.parallel_dict.items():
                input_name = "/".join(name.split("/")[:-1])
                output_shapes.update(layer.compute_output_shape(input_shape[input_name]))

            return output_shapes

        return super().compute_call_output_shape(input_shape)

    @classmethod
    def from_config(cls, config, custom_objects=None, **kwargs):
        inputs, config = ParallelBlock.parse_config(config, custom_objects)
        paralel_block = ParallelBlock(inputs, **config)
        paralel_block.__class__ = cls
        return paralel_block

    def get_config(self):
        config = super().get_config()
        return config


def create_expert(expert_block: Block, name: str) -> Tuple[str, TabularBlock]:
    """Creates an expert from a block

    Parameters
    ----------
    expert_block : Block
        Expert Block
    name : str
        Expert name

    Returns
    -------
    Tuple[str, TabularBlock]
        Tuple with the expert name and block
    """
    return name, expert_block.copy().as_tabular(name)


def PLEBlock(
    num_layers: int,
    outputs: Union[List[str], List[PredictionTask], ParallelPredictionBlock],
    expert_block: Union[Block, tf.keras.layers.Layer],
    num_task_experts: int = 1,
    num_shared_experts: int = 1,
    gate_block: Optional[Block] = None,
    gate_softmax_temperature: float = 1.0,
    enable_gate_weights_metrics: bool = False,
    name: Optional[str] = None,
    **kwargs,
):
    """Implements the Progressive Layered Extraction (PLE) model from [1],
    by stacking CGC blocks (CGCBlock).

    References
    ----------
    [1] Tang, Hongyan, et al. "Progressive layered extraction (ple): A novel multi-task
    learning (mtl) model for personalized recommendations."
    Fourteenth ACM Conference on Recommender Systems. 2020.

    Parameters
    ----------
    num_layers : int
        Number of stacked CGC blocks
    outputs : Union[List[str], List[PredictionTask], ParallelPredictionBlock]
        Names of the tasks or PredictionTask/ParallelPredictionBlock objects from
        which we can extract the task names
    expert_block : Union[Block, tf.keras.layers.Layer]
        Block that will be used for the experts
    num_task_experts : int, optional
        Number of task-specific experts, by default 1
    num_shared_experts : int, optional
        Number of shared experts for tasks, by default 1
    gate_block : Optional[Block], optional
        Optional block than can make the Gate mode powerful in converting
        the inputs into expert weights for averaging, by default None
    gate_softmax_temperature : float, optional
        Temperature of the softmax used by the Gates for getting
        weights for the average. Temperature can be used to smooth
        the weights distribution, and is by default 1.0.
    enable_gate_weights_metrics : bool, optional
        Enables logging the average gate weights on experts
    name : Optional[str], optional
        Name of the PLE block, by default None

    Returns
    -------
    SequentialBlock
        Returns the PLE block
    """
    cgc_blocks = []

    for i in range(num_layers):
        cgc_block = CGCBlock(
            outputs=outputs,
            expert_block=expert_block,
            num_task_experts=num_task_experts,
            num_shared_experts=num_shared_experts,
            add_shared_gate=(i < num_layers - 1),
            gate_block=gate_block,
            gate_softmax_temperature=gate_softmax_temperature,
            enable_gate_weights_metrics=enable_gate_weights_metrics,
            name=f"cgc_block_{i}",
            **kwargs,
        )
        cgc_blocks.append(cgc_block)

    cgc_blocks = SequentialBlock(*cgc_blocks, name=name)
    return cgc_blocks
